from pathlib import Path
from typing import Any
from typing import Dict
from typing import Optional

from semgrep.semgrep_types import PatternId
from semgrep.semgrep_types import Range


class PatternMatch:
    """
    Encapsulates a section of code that matches a single pattern
    """

    def __init__(self, raw_json: Dict[str, Any]) -> None:
        self._raw_json = raw_json
        self._metavariable_values: Optional[Dict[str, str]] = None

    @property
    def rule_index(self) -> int:
        return int(self._raw_json["check_id"].split(".")[0])

    @property
    def id(self) -> PatternId:
        return PatternId(".".join(self._raw_json["check_id"].split(".")[1:]))

    @property
    def path(self) -> Path:
        return Path(self._raw_json["path"])

    @property
    def metavariables(self) -> Dict[str, Any]:
        return self.extra.get("metavars", {})

    @property
    def extra(self) -> Dict[str, Any]:
        return self._raw_json["extra"]

    @property
    def metavariable_uids(self) -> Dict[str, Any]:
        def _get_uid(metavariable_data: Any) -> Any:
            try:
                return metavariable_data["unique_id"]["sid"]
            except KeyError:
                try:
                    return metavariable_data["unique_id"]["md5sum"]
                except KeyError:
                    return None

        return {
            metavariable: _get_uid(data)
            for metavariable, data in self.metavariables.items()
        }

    @property
    def range(self) -> Range:
        return Range(
            self._raw_json["start"]["offset"],
            self._raw_json["end"]["offset"],
            self.metavariable_uids,
        )

    @property
    def start(self) -> Dict[str, Any]:
        # https://docs.r2c.dev/en/latest/api/output.html does not support offset at the moment
        start = dict(self._raw_json["start"])
        if "offset" in start:
            del start["offset"]
        return start

    @property
    def end(self) -> Dict[str, Any]:
        # https://docs.r2c.dev/en/latest/api/output.html does not support offset at the moment
        end = dict(self._raw_json["end"])
        if "offset" in end:
            del end["offset"]
        return end

    def _read_metavariable_values(self) -> Dict[str, str]:
        """
        Read self.path and lookup all values of metavariables in self.metavariables
        """
        result = {}

        # open path and ignore non-utf8 bytes. https://stackoverflow.com/a/56441652
        with open(self.path, errors="replace") as fd:
            for metavariable, metavariable_data in self.metavariables.items():
                # Offsets are start inclusive and end exclusive
                start_offset = metavariable_data["start"]["offset"]
                end_offset = metavariable_data["end"]["offset"]
                length = end_offset - start_offset

                fd.seek(start_offset)
                result[metavariable] = fd.read(length)

        return result

    def get_metavariable_value(self, metavariable: str) -> str:
        """
        Use metavariable's start and end to read into the file to find what the
        metavariable in this pattern match maps to in the file

        Assumes METAVARIABLE is a key in self.metavariables
        """
        if self._metavariable_values is None:
            self._metavariable_values = self._read_metavariable_values()

        return self._metavariable_values[metavariable]

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} range={self.range}>"



import pytest


def xfail_repo(repo, *, reason=""):
    return pytest.param(repo, marks=pytest.mark.xfail(reason=reason, strict=True))


ALL_LANGUAGES = ["ALL"]


PASSING_REPOS = [
    {"repo": "https://github.com/coinbase/bifrost", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/coinbase/bip38", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/coinbase/btcexport", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/coinbase/coinbase-android-sdk",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/coinbase/coinbase-android-sdk-example",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/coinbase/coinbase-bitmonet-sdk",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/coinbase/coinbase-commerce-node",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/coinbase/coinbase-commerce-python",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/coinbase/coinbase-exchange-node",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/coinbase/coinbase-java", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/coinbase/coinbase-javascript-sdk",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/coinbase/coinbase-node", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/coinbase/coinbase-python", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/coinbase/coinbase-tip-discourse",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/coinbase/dexter", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/coinbase/fenrir", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/coinbase/node-process-lock",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/coinbase/odin", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/coinbase/protoc-gen-rbi", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/coinbase/pwnbot", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/coinbase/rosetta-cli", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/coinbase/rosetta-sdk-go", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/coinbase/self-service-iam",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/coinbase/solidity-workshop",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/coinbase/step", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/coinbase/step-asg-deployer",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/coinbase/watchdog", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/AskBox", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/dropbox/Developer-Samples",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/dropbox/changes-artifacts",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/dropbox/changes-client", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/dropbox/changes-lxc-wrapper",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/dropbox/changes-mesos-framework",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/dropbox/dbx-unittest2pytest",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/dropbox/dbxcli", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/dropbox/dropbox-api-content-hasher",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/dropbox/dropbox-api-v2-repl",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/dropbox/dropbox-sdk-go-unofficial",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/dropbox/dropbox-sdk-java", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/dropbox-sdk-js", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/dropbox/dropbox-sdk-python",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/dropbox/dropbox_hook", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/emmer", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/dropbox/firebase-dropbox-oauth",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/dropbox/git-rbr", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/PyHive", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/goebpf", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/goprotoc", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/grallama-panel", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/groupy", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/hermes", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/hocrux", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/hydra", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/llama", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/llama-archive", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/load_management", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/mdwebhook", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/merou", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/dropbox/mypy-PyCharm-plugin",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/dropbox/mypy-protobuf", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/nautilus-dropbox", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/nsot", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/othw", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/dropbox/pem-converter-maven-plugin",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/dropbox/pep8squad", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/dropbox/presto-kafka-connector",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/dropbox/puppet_run", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/pyannotate", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/pygerduty", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/pynsot", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/dropbox/pytest-call-tracer",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/dropbox/pytest-flakefinder",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/dropbox/python-invariant", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/revision-browser", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/rules_node", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/securitybot", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/spookify", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/sqlalchemy-stubs", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/stone", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/stopwatch", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/strongpoc", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/whitegold", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/ykfipsconf", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/zinger", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/returntocorp/badwords", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/returntocorp/bento-report",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/returntocorp/buffer-rule-tests",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/returntocorp/check-docs", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/returntocorp/cli", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/returntocorp/flake8-click",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/returntocorp/flake8-flask",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/returntocorp/flake8-requests",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/returntocorp/inputset-generator",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/returntocorp/semgrep-action",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/returntocorp/semgrep-rules",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/seemoo-lab/opendrop", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/lightstep/lightstep-tracer-python",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/draios/sysdig-inspect", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/getsentry/sentry-python", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/signalapp/signal-webrtc-ios",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/secdev/scapy", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/apache/airflow", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/preset-io/elasticsearch-dbapi",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/apache/libcloud", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/keybase/pykeybasebot", "languages": ALL_LANGUAGES},
    {
        "repo": "https://gitbox.apache.org/repos/asf/cassandra",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/keybase/python-triplesec", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/psycopg/psycopg2", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/preset-io/flask-jwt-extended",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/vstinner/pyperf", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/mysql/mysql-connector-python",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/Netflix/lemur", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/mpirnat/lets-be-bad-guys", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/JasonHinds13/hackable", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/ab-smith/gruyere", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/digininja/vuLnDAP", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/godropbox", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/trapperkeeper", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/lodash/lodash", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/home-assistant/home-assistant",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/0c34/govwa", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/coinbase/bittip", "languages": ALL_LANGUAGES},
    {
        "repo": "https://github.com/coinbase/multisig-tool",
        "languages": ALL_LANGUAGES,
    },
    {
        "repo": "https://github.com/jekyll/jekyll",
        "languages": ALL_LANGUAGES,
    },
    {"repo": "https://github.com/OWASP/NodeGoat", "languages": ALL_LANGUAGES},
    {"repo": "https://github.com/dropbox/questions", "languages": ALL_LANGUAGES},
    # Excluding cause of https://github.com/returntocorp/semgrep/issues/2613
    {
        "repo": "https://github.com/zulip/zulip",
        "languages": ALL_LANGUAGES,
        "excludes": ["frontend_tests/puppeteer_lib/common.ts"],
    },
]

FAILING_REPOS = [
    # this fails in some CI, but works locally, weird
    #    xfail_repo(
    #        {
    #            "repo": "https://github.com/we45/Vulnerable-Flask-App",
    #            "languages": ALL_LANGUAGES,
    #        },
    #        reason="MatchTimeout error but happens only in CI",
    #    ),
    xfail_repo(
        {"repo": "https://github.com/rails/rails", "languages": ["ruby"]},
        reason="https://github.com/tree-sitter/tree-sitter-ruby/issues/167",
    ),
    xfail_repo(
        {
            "repo": "https://github.com/highcharts/highcharts",
            "languages": ALL_LANGUAGES,
        },
        reason="javascript parse error, probably due to '?.01' regression in minified file",
    ),
    xfail_repo(
        {"repo": "https://github.com/coinbase/gtt-ui", "languages": ALL_LANGUAGES},
        reason="javascript parse error, probably due to ?. regression",
    ),
    xfail_repo(
        {"repo": "https://github.com/DevSlop/Pixi", "languages": ALL_LANGUAGES},
        reason="javascript parse error, probably due to ?. regression",
    ),
    xfail_repo(
        {
            "repo": "https://github.com/dropbox/incubator-superset-internal",
            "languages": ALL_LANGUAGES,
        },
        reason="javascript parse error",
    ),
    xfail_repo(
        {
            "repo": "https://github.com/coinbase/react-coinbase-commerce",
            "languages": ALL_LANGUAGES,
        },
        reason="react",
    ),
    xfail_repo(
        {
            "repo": "https://github.com/dropbox/DropboxBusinessScripts",
            "languages": ALL_LANGUAGES,
        },
        reason="https://github.com/returntocorp/semgrep/issues/1498",
    ),
    xfail_repo(
        {
            "repo": "https://github.com/dropbox/dbx_build_tools",
            "languages": ALL_LANGUAGES,
        },
        reason="go parse errors",
    ),
    xfail_repo(
        {"repo": "https://github.com/dropbox/hackpad", "languages": ALL_LANGUAGES},
        reason="https://www.oreilly.com/library/view/javascript-the-definitive/9781449393854/ch11s06.html",
    ),
    xfail_repo(
        {
            "repo": "https://github.com/dropbox/notouch",
            "languages": ALL_LANGUAGES,
        },
        reason="indented comment on last line python",
    ),
    xfail_repo(
        {
            "repo": "https://github.com/rapid7/metasploit-framework",
            "languages": ["ruby"],
        }
    ),
    xfail_repo({"repo": "https://github.com/Homebrew/brew", "languages": ["ruby"]}),
    xfail_repo({"repo": "https://github.com/gitlabhq/gitlabhq", "languages": ["ruby"]}),
    #    xfail_repo(
    #        {
    #            "repo": "https://github.com/coinbase/coinbase-pro-node",
    #            "languages": ["javascript", "typescript"],
    #        },
    #        reason=("Failure to parse typescript"),
    #    ),
    xfail_repo(
        {
            "repo": "https://github.com/bkimminich/juice-shop",
            "languages": ["javascript", "typescript"],
        },
        reason=("Failure to parse typescript"),
    ),
    xfail_repo(
        {"repo": "https://github.com/dropbox/changes", "languages": ALL_LANGUAGES},
        reason="react",
    ),
    xfail_repo(
        {"repo": "https://github.com/dropbox/pyston-perf", "languages": ALL_LANGUAGES},
        reason="templates",
    ),
    xfail_repo(
        {
            "repo": "https://github.com/opensourceactivismtech/call-power",
            "languages": ALL_LANGUAGES,
        },
        reason="templates",
    ),
]

# Access this list with the `public_repo_url` fixture argument.
ALL_REPOS = FAILING_REPOS + PASSING_REPOS



import pytest


@pytest.mark.parametrize(
    "rule,target",
    [
        (
            "rules/message_interpolation/pattern-inside.yaml",
            "message_interpolation/pattern_inside_basic.py",
        ),
        (
            "rules/message_interpolation/pattern-inside.yaml",
            "message_interpolation/pattern_inside_complex.py",
        ),
        # pattern-not-inside is not currently interpolated. These tests make sure
        # something doesn't break accidentally.
        (
            "rules/message_interpolation/pattern-not-inside.yaml",
            "message_interpolation/pattern_not_inside_basic.py",
        ),
        (
            "rules/message_interpolation/pattern-not-inside.yaml",
            "message_interpolation/pattern_not_inside_complex.py",
        ),
        (
            "rules/message_interpolation/pattern-either.yaml",
            "message_interpolation/pattern_either_basic.py",
        ),
        (
            "rules/message_interpolation/multi-pattern-inside.yaml",
            "message_interpolation/multi_pattern_inside.py",
        ),
        (
            "rules/message_interpolation/multi-pattern-inside.yaml",
            "message_interpolation/multi_pattern_inside_nested.py",
        ),
    ],
)
def test_message_interpolation(run_semgrep_in_tmp, snapshot, rule, target):
    snapshot.assert_match(
        run_semgrep_in_tmp(rule, target_name=target),
        "results.json",
    )



"""
For each directory containing YAML rules, run those rules on the file in the same directory with the same name but different extension.
E.g. eqeq.yaml runs on eqeq.py.
Validate that the output is annotated in the source file with by looking for a comment like:

 ```
 # ruleid:eqeq-is-bad
 ```
 On the preceeding line.

 """
import argparse
import collections
import functools
import json
import logging
import multiprocessing
import sys
import tarfile
from itertools import product
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Set
from typing import Tuple

from semgrep.constants import BREAK_LINE
from semgrep.semgrep_main import invoke_semgrep
from semgrep.util import is_config_suffix
from semgrep.util import is_config_test_suffix
from semgrep.util import partition

logger = logging.getLogger(__name__)

SAVE_TEST_OUTPUT_JSON = "semgrep_runs_output.json"
SAVE_TEST_OUTPUT_TAR = "semgrep_runs_output.tar.gz"

COMMENT_SYNTAXES = (("#", "\n"), ("//", "\n"), ("<!--", "-->"), ("(*", "*)"))
SPACE_OR_NO_SPACE = ("", " ")
TODORULEID = "todoruleid"
RULEID = "ruleid"
TODOOK = "todook"
OK = "ok"


def _remove_ending_comments(rule: str) -> str:
    for _, end in COMMENT_SYNTAXES:
        rule = rule.strip() if end == "\n" else rule.strip().replace(end, "")
    return rule


def normalize_rule_ids(line: str) -> Set[str]:
    """
    given a line like `     # ruleid:foobar`
    or `      // ruleid:foobar`
    return `foobar`
    """
    _, rules_text = line.strip().split(":")
    rules_text = rules_text.strip()
    rules = rules_text.split(",")
    # remove comment ends for non-newline comment syntaxes
    rules_clean = map(lambda rule: _remove_ending_comments(rule), rules)
    return set(filter(None, [rule.strip() for rule in rules_clean]))


def compute_confusion_matrix(
    reported: Set[Any], expected: Set[Any], oked: Set[Any]
) -> Tuple[int, int, int, int]:
    true_positives = len(expected.intersection(reported))
    false_positives = len(reported - expected)
    true_negatives = len(oked)
    false_negatives = len(expected - reported)

    return (true_positives, true_negatives, false_positives, false_negatives)


def _test_compute_confusion_matrix() -> None:
    tp, tn, fp, fn = compute_confusion_matrix(set([1, 2, 3, 4]), set([1]), set())
    assert tp == 1
    assert tn == 0
    assert fp == 3
    assert fn == 0

    tp, tn, fp, fn = compute_confusion_matrix(
        set([1, 2, 3, 4]), set([1, 2, 3, 4]), set([1])
    )
    assert tp == 4
    assert tn == 1
    assert fp == 0
    assert fn == 0

    tp, tn, fp, fn = compute_confusion_matrix(
        set([2, 3]), set([1, 2, 3, 4]), set([7, 8])
    )
    assert tp == 2
    assert tn == 2
    assert fp == 0
    assert fn == 2


def _annotations(annotation: str) -> Set[str]:
    # returns something like: {"#ruleid:", "# ruleid:", "//ruleid:", ...}
    return {
        f"{comment_syntax[0]}{space}{annotation}"
        for comment_syntax, space in product(COMMENT_SYNTAXES, SPACE_OR_NO_SPACE)
    }


def line_has_todo_rule(line: str) -> bool:
    todo_rule_annotations = _annotations(TODORULEID)
    return any(annotation in line for annotation in todo_rule_annotations)


def line_has_rule(line: str) -> bool:
    rule_annotations = _annotations(RULEID)
    return any(annotation in line for annotation in rule_annotations)


def line_has_ok(line: str) -> bool:
    rule_annotations = _annotations(OK)
    return any(annotation in line for annotation in rule_annotations)


def line_has_todo_ok(line: str) -> bool:
    rule_annotations = _annotations(TODOOK)
    return any(annotation in line for annotation in rule_annotations)


def score_output_json(
    json_out: Dict[str, Any], test_files: List[Path], ignore_todo: bool
) -> Tuple[Dict[str, List[int]], Dict[str, Dict[str, Any]], int]:
    ruleid_lines: Dict[str, Dict[str, List[int]]] = collections.defaultdict(
        lambda: collections.defaultdict(list)
    )
    ok_lines: Dict[str, Dict[str, List[int]]] = collections.defaultdict(
        lambda: collections.defaultdict(list)
    )
    reported_lines: Dict[str, Dict[str, List[int]]] = collections.defaultdict(
        lambda: collections.defaultdict(list)
    )
    todo_ok_lines: Dict[str, Dict[str, List[int]]] = collections.defaultdict(
        lambda: collections.defaultdict(list)
    )
    score_by_checkid: Dict[str, List[int]] = collections.defaultdict(
        lambda: [0, 0, 0, 0]
    )
    matches_by_check_id: Dict[str, Dict[str, Any]] = collections.defaultdict(dict)
    num_todo = 0

    for test_file in test_files:
        test_file_resolved = str(test_file.resolve())
        all_lines = test_file.read_text().split("\n")
        for i, line in enumerate(all_lines):
            # +1 because we are 0 based and semgrep output is not, plus skip the comment line
            effective_line_num = i + 2

            rule_in_line = line_has_rule(line)
            ok_in_line = line_has_ok(line)
            todo_rule_in_line = line_has_todo_rule(line)
            todo_ok_in_line = line_has_todo_ok(line)
            num_todo += int(todo_rule_in_line) + int(todo_ok_in_line)

            try:
                if (not ignore_todo and todo_rule_in_line) or rule_in_line:
                    rule_ids = normalize_rule_ids(line)
                    for rule_id in rule_ids:
                        ruleid_lines[test_file_resolved][rule_id].append(
                            effective_line_num
                        )
                if (not ignore_todo and todo_rule_in_line) or ok_in_line:
                    rule_ids = normalize_rule_ids(line)
                    for rule_id in rule_ids:
                        ok_lines[test_file_resolved][rule_id].append(effective_line_num)
                if ignore_todo and todo_ok_in_line:
                    rule_ids = normalize_rule_ids(line)
                    for rule_id in rule_ids:
                        todo_ok_lines[test_file_resolved][rule_id].append(
                            effective_line_num
                        )
            except ValueError:  # comment looked like a test annotation but couldn't parse
                logger.warning(
                    f"Could not parse {line} as a test annotation in file {test_file_resolved}. Skipping this line"
                )
                continue

    for result in json_out["results"]:
        reported_lines[str(Path(result["path"]).resolve())][result["check_id"]].append(
            int(result["start"]["line"])
        )

    reported_ids = {result["check_id"] for result in json_out["results"]}
    test_ids = {
        check_id for _, ruleid in ruleid_lines.items() for check_id in ruleid.keys()
    }
    test_ids |= {check_id for _, ok in ok_lines.items() for check_id in ok.keys()}
    if reported_ids and test_ids.symmetric_difference(reported_ids):
        raise Exception(
            f"found mismatch between test and result ids - test={test_ids} result={reported_ids}"
        )

    def join_keys(a: Dict[str, Any], b: Dict[str, Any]) -> Set[str]:
        return set(a.keys()).union(set(b.keys()))

    for file_path in join_keys(ruleid_lines, reported_lines):
        for check_id in join_keys(ruleid_lines[file_path], reported_lines[file_path]):
            all_reported = set(reported_lines[file_path][check_id])
            expected = set(ruleid_lines[file_path][check_id])
            oked = set(ok_lines[file_path][check_id])
            todo_oked = set(todo_ok_lines[file_path][check_id])

            reported_oked_lines = oked.intersection(all_reported)
            if reported_oked_lines:
                raise Exception(
                    f"found results on ok'ed lines - lines={reported_oked_lines} path={file_path}"
                )

            reported = all_reported - todo_oked

            new_cm = compute_confusion_matrix(reported, expected, oked)
            matches_by_check_id[check_id][file_path] = {
                "expected_lines": sorted(expected),
                "reported_lines": sorted(reported),
            }
            old_cm = score_by_checkid[check_id]
            score_by_checkid[check_id] = [
                old_cm[i] + new_cm[i] for i in range(len(new_cm))
            ]

    return (score_by_checkid, matches_by_check_id, num_todo)


def generate_confusion_string(check_results: Mapping[str, Any]) -> str:
    confusion_tp = f"TP: {check_results['tp']}"
    confusion_tn = f"TN: {check_results['tn']}"
    confusion_fp = f"FP: {check_results['fp']}"
    confusion_fn = f"FN: {check_results['fn']}"
    return f"{confusion_tp} {confusion_tn} {confusion_fp} {confusion_fn}"


def generate_check_output_line(check_id: str, check_results: Mapping[str, Any]) -> str:
    status = "✔" if check_results["passed"] else "✖"
    return f"\t{status} {check_id.ljust(60)} {generate_confusion_string(check_results)}"


def generate_matches_line(check_results: Mapping[str, Any]) -> str:
    def _generate_line(test_file: Any, matches: Mapping[str, Any]) -> str:
        test = f"test: {test_file}"
        expected = f"expected lines: {matches['expected_lines']}"
        reported = f"reported lines: {matches['reported_lines']}"
        return f"{test}, {expected}, {reported}"

    return "\t" + "\t\n".join(
        _generate_line(test_file, matches)
        for test_file, matches in check_results["matches"].items()
    )


def invoke_semgrep_multi(
    config: Path, targets: List[Path], **kwargs: Any
) -> Tuple[Path, Optional[Exception], Any]:
    try:
        output = invoke_semgrep(config, targets, **kwargs)
    except Exception as error:
        return (config, error, {})
    else:
        return (config, None, output)


def relatively_eq(parent1: Path, child1: Path, parent2: Path, child2: Path) -> bool:
    def remove_all_suffixes(p: Path) -> Path:
        result = p.with_suffix("")
        while result != result.with_suffix(""):
            result = result.with_suffix("")
        return result

    rel1 = remove_all_suffixes(child1.relative_to(parent1))
    rel2 = remove_all_suffixes(child2.relative_to(parent2))
    return rel1 == rel2


def get_config_filenames(original_config: Path) -> List[Path]:
    configs = list(original_config.rglob("*"))
    return [
        config
        for config in configs
        if is_config_suffix(config)
        and not config.name.startswith(".")
        and not config.parent.name.startswith(".")
    ]


def get_config_test_filenames(
    original_config: Path, configs: List[Path], original_target: Path
) -> Dict[Path, List[Path]]:
    targets = list(original_target.rglob("*"))

    def target_matches_config(target: Path, config: Path) -> bool:
        correct_suffix = is_config_test_suffix(target) or not is_config_suffix(target)
        return (
            relatively_eq(original_target, target, original_config, config)
            and target.is_file()
            and correct_suffix
        )

    return {
        config: [target for target in targets if target_matches_config(target, config)]
        for config in configs
    }


def generate_file_pairs(
    target: Path,
    config: Path,
    ignore_todo: bool,
    strict: bool,
    unsafe: bool,
    json_output: bool,
    save_test_output_tar: bool = True,
    optimizations: str = "none",
) -> None:
    config_filenames = get_config_filenames(config)
    config_test_filenames = get_config_test_filenames(config, config_filenames, target)
    config_with_tests, config_without_tests = partition(
        lambda c: c[1], config_test_filenames.items()
    )
    config_missing_tests_output = [str(c[0]) for c in config_without_tests]

    invoke_semgrep_fn = functools.partial(
        invoke_semgrep_multi,
        no_git_ignore=True,
        no_rewrite_rule_ids=True,
        strict=strict,
        dangerously_allow_arbitrary_code_execution_from_rules=unsafe,
        optimizations=optimizations,
    )
    with multiprocessing.Pool(multiprocessing.cpu_count()) as pool:
        results = pool.starmap(invoke_semgrep_fn, config_with_tests)

    config_with_errors, config_without_errors = partition(lambda r: r[1], results)
    config_with_errors_output = [
        {"filename": str(filename), "error": str(error), "output": output}
        for filename, error, output in config_with_errors
    ]

    tested = {
        filename: score_output_json(
            output, config_test_filenames[filename], ignore_todo
        )
        for filename, _, output in config_without_errors
    }

    results_output: Mapping[str, Mapping[str, Any]] = {
        str(filename): {
            "todo": todo,
            "checks": {
                check_id: {
                    "tp": tp,
                    "tn": tn,
                    "fp": fp,
                    "fn": fn,
                    "passed": (fp == 0) and (fn == 0),
                    "matches": matches[check_id],
                }
                for check_id, (tp, tn, fp, fn) in output.items()
            },
        }
        for filename, (output, matches, todo) in tested.items()
    }

    output = {
        "config_missing_tests": config_missing_tests_output,
        "config_with_errors": config_with_errors_output,
        "results": results_output,
    }

    strict_error = bool(config_with_errors_output) and strict
    any_failures = any(
        not check_results["passed"]
        for file_results in results_output.values()
        for check_results in file_results["checks"].values()
    )
    exit_code = int(strict_error or any_failures)

    if json_output:
        print(json.dumps(output, indent=4, separators=(",", ": ")))
        sys.exit(exit_code)

    # save the results to json file and tar the file to upload as github artifact.
    if save_test_output_tar:
        list_to_output = []
        with open(SAVE_TEST_OUTPUT_JSON, "w") as f:
            for tup in results:
                true_result = tup[2]
                list_to_output.append(true_result)
            f.write(json.dumps(list_to_output, indent=4, separators=(",", ":")))

        with tarfile.open(SAVE_TEST_OUTPUT_TAR, "w:gz") as tar:
            tar.add(SAVE_TEST_OUTPUT_JSON)

    if config_missing_tests_output:
        print("The following config files are missing tests:")
        print("\t" + "\n\t".join(config_missing_tests_output))

    if config_with_errors_output:
        print("The following config files produced errors:")
        print(
            "\t"
            + "\n\t".join(
                f"{c['filename']}: {c['error']}" for c in config_with_errors_output
            )
        )

    # Place failed and TODO tests at the bottom for higher visibility
    passed_results_first = collections.OrderedDict(
        sorted(
            results_output.items(),
            key=lambda t: any(
                not c["passed"] or t[1]["todo"] for c in t[1]["checks"].values()
            ),
        )
    )

    print(f"{len(tested)} yaml files tested")
    print("check id scoring:")
    print(BREAK_LINE)

    totals: Dict[str, Any] = collections.defaultdict(int)

    for filename, rr in passed_results_first.items():
        print(f"(TODO: {rr['todo']}) {filename}")
        for check_id, check_results in sorted(rr["checks"].items()):
            print(generate_check_output_line(check_id, check_results))
            if not check_results["passed"]:
                print(generate_matches_line(check_results))
            for confusion in ["tp", "tn", "fp", "fn"]:
                totals[confusion] += check_results[confusion]

    print(BREAK_LINE)
    print(f"final confusion matrix: {generate_confusion_string(totals)}")
    print(BREAK_LINE)

    sys.exit(exit_code)


def test_main(args: argparse.Namespace) -> None:
    _test_compute_confusion_matrix()

    if len(args.target) != 1:
        raise Exception("only one target directory allowed for tests")
    target = Path(args.target[0])

    if args.config:
        if len(args.config) != 1:
            raise Exception("only one config directory allowed for tests")
        config = Path(args.config[0])
    else:
        config = target

    generate_file_pairs(
        target,
        config,
        args.test_ignore_todo,
        args.strict,
        args.dangerously_allow_arbitrary_code_execution_from_rules,
        args.json,
        args.save_test_output_tar,
        args.optimizations,
    )



import collections
from pathlib import Path
from typing import Any
from typing import cast
from typing import Dict
from typing import Set

from semgrep.target_manager_extensions import ext_to_langs
from semgrep.target_manager_extensions import FileExtension
from semgrep.target_manager_extensions import Language


def make_target_stats(all_targets: Set[Path]) -> Dict[str, Any]:
    """
    Given a set of all the targets semgrep ran on, compute stats about
    the distribution of files involved.

    You could argue that we could compute the languages from the extensions later,
    which is true. But since semgrep's mapping might change by version, I think it's
    helpful to also record what language semgrep used.
    """
    # example:
    # languages = { 'python': 105, javascript: '308', 'generic': 20 }
    # extensions = { 'py': 10, 'pyi': 8 }
    languages: Dict[Language, int] = collections.defaultdict(int)
    extensions: Dict[FileExtension, int] = collections.defaultdict(int)
    for path in all_targets:
        suffix: FileExtension = cast(FileExtension, path.suffix)
        extensions[suffix] += 1
        # an extension could map to multiple languages; just take the first one
        lang = sorted(ext_to_langs(suffix))[0]
        languages[lang] += 1

    return {
        # explicit convert from a defaultdict to a dict to make sure nothing else
        # accidentally inserts entries
        "extensions": dict(extensions),
        "languages": dict(languages),
    }


def make_loc_stats(all_targets: Set[Path]) -> Dict[str, Any]:
    return {"by_extension": count_lines_by_path_extension(all_targets)}


def count_lines_in_file(p: Path) -> int:
    """
    Return number of lines in the path. We're assuming this path is an openable file,
    not a symlink or a directory -- it should have already been checked when it was
    generated in the TargetManager
    """
    # https://stackoverflow.com/a/37600991
    return sum(1 for i in open(p, "rb"))


def count_lines_by_path_extension(paths: Set[Path]) -> Dict[FileExtension, int]:
    all_counts = {p: count_lines_in_file(p) for p in paths}
    by_extension: Dict[FileExtension, int] = collections.defaultdict(int)
    for p, count in all_counts.items():
        by_extension[cast(FileExtension, p.suffix)] += count
    return dict(by_extension)



from enum import Enum
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import attr
from colorama import Fore

from semgrep.rule_lang import Position
from semgrep.rule_lang import SourceTracker
from semgrep.rule_lang import Span
from semgrep.util import with_color

OK_EXIT_CODE = 0
FINDINGS_EXIT_CODE = 1
FATAL_EXIT_CODE = 2
INVALID_CODE_EXIT_CODE = 3
INVALID_PATTERN_EXIT_CODE = 4
UNPARSEABLE_YAML_EXIT_CODE = 5
NEED_ARBITRARY_CODE_EXEC_EXIT_CODE = 6
MISSING_CONFIG_EXIT_CODE = 7
INVALID_LANGUAGE_EXIT_CODE = 8
MATCH_TIMEOUT_EXIT_CODE = 9
MATCH_MAX_MEMORY_EXIT_CODE = 10
LEXICAL_ERROR_EXIT_CODE = 11
TOO_MANY_MATCHES_EXIT_CODE = 12


class Level(Enum):
    ERROR = 4  # Always an error
    WARN = 3  # Only an error if "strict" is set


class SemgrepError(Exception):
    """
    Parent class of all exceptions we anticipate in Semgrep commands

    All Semgrep Exceptions are caught and their error messages
    are displayed to the user.

    For pretty-printing, exceptions should override `__str__`.
    """

    def __init__(
        self, *args: object, code: int = FATAL_EXIT_CODE, level: Level = Level.ERROR
    ) -> None:
        self.code = code
        self.level = level

        super().__init__(*args)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.__class__.__name__,
            "code": self.code,
            "level": self.level.name.lower(),
            **self.to_dict_base(),
        }

    def to_dict_base(self) -> Dict[str, Any]:
        """
        Default implementation. Subclasses should override to provide custom information.
        All values returned must be JSON serializable.
        """
        return {"message": str(self)}


class SemgrepInternalError(Exception):
    """
    Parent class of internal semgrep exceptions that should be handled internally and converted into `SemgrepError`s

    Classes that inherit from SemgrepInternalError should begin with `_`
    """

    pass


class UnknownOperatorError(SemgrepError):
    pass


@attr.s(auto_attribs=True, frozen=True)
class FilesNotFoundError(SemgrepError):
    level = Level.ERROR
    code = FATAL_EXIT_CODE
    paths: Sequence[Path]

    def __str__(self) -> str:
        lines = (f"File not found: {pathname}" for pathname in self.paths)
        return "\n".join(lines)


def span_list_to_tuple(spans: List[Span]) -> Tuple[Span, ...]:
    """
    Helper converter so mypy can track that we are converting
    from list of spans to tuple of spans
    """
    return tuple(spans)


@attr.s(auto_attribs=True, eq=True, frozen=True)
class ErrorWithSpan(SemgrepError):
    """
    In general, you should not be constructing ErrorWithSpan directly, and instead be constructing a subclass
    that sets the code.

    Error which will print context from the Span. You should provide the most specific span possible,
    eg. if the error is an invalid key, provide exactly the span for that key. You can then expand what's printed
    with span.with_context(...). Conversely, if you don't want to display the entire span, you can use `span.truncate`

    The __str__ method produces the pretty-printed error.
    Here is what the generated error will look like:

        <level>: <short_msg>
          --> <span.filename>:<span.start.line>
        1 | rules:
        2 |   - id: eqeq-is-bad
        3 |     pattern-inside: foo(...)
          |     ^^^^^^^^^^^^^^
        4 |     patterns:
        5 |       - pattern-not: 1 == 1
        = help: <help>
        <long_msg>

    :param short_msg: 1 or 2 word description of the problem (eg. missing key)
    :param level: How bad is the problem? error,warn, etc.
    :param spans: A list of spans to display for context.
    :help help: An optional hint about how to fix the problem
    :cause cause: The underlying exception
    """

    short_msg: str = attr.ib()
    long_msg: Optional[str] = attr.ib()
    spans: List[Span] = attr.ib(converter=span_list_to_tuple)
    help: Optional[str] = attr.ib(default=None)

    def __attrs_post_init__(self) -> None:
        if not hasattr(self, "code"):
            raise ValueError("Inheritors of SemgrepError must define an exit code")

        if not hasattr(self, "level"):
            raise ValueError("Inheritors of SemgrepError must define a level")

    def to_dict_base(self) -> Dict[str, Any]:
        base = dict(
            short_msg=self.short_msg,
            long_msg=self.long_msg,
            level=self.level.name.lower(),
            spans=[attr.asdict(s) for s in self.spans],
        )
        # otherwise, we end up with `help: null` in JSON
        if self.help:
            base["help"] = self.help
        return base

    @staticmethod
    def _line_number_width(span: Span) -> int:
        return len(str((span.context_end or span.end).line)) + 1

    @staticmethod
    def _format_line_number(span: Span, line_number: Optional[int]) -> str:
        """
        Produce a string like:
        ` 10 |`

        The amount of padding is set for printing within `span` (so it handles up to `context_end.line`)
        """
        # line numbers are 0 indexed
        width = ErrorWithSpan._line_number_width(span)
        if line_number is not None:
            base_str = str(line_number)
            assert len(base_str) < width
            return with_color(Fore.LIGHTBLUE_EX, base_str.ljust(width) + "| ")
        else:
            return with_color(Fore.LIGHTBLUE_EX, "".ljust(width) + "| ")

    def _format_code_segment(
        self, start: Position, end: Position, source: List[str], part_of_span: Span
    ) -> List[str]:
        """
        Line by line output for a snippet of code from `start_line` to `end_line`
        Each line will be annotated with a line number, properly spaced according to
        the highest line number required to render `span`

        :param start: start position
        :param end: end position

        :returns A list of strings, suitable to be combined with `'\n'.join(...)`
        eg:
        List[
            "5  | def my_func():",
            "6  |   return True"
        ]
        """
        # -1 because positions are 1-indexed
        code_segment = source[start.line - 1 : end.line]
        snippet = []
        for line_num, line in zip(range(start.line, end.line + 1), code_segment):
            snippet.append(f"{self._format_line_number(part_of_span, line_num)}{line}")
        return snippet

    def __str__(self) -> str:
        """
        Format this exception into a pretty string with context and color
        """
        header = f"{with_color(Fore.RED, 'semgrep ' + self.level.name.lower())}: {self.short_msg}"
        snippets = []
        for span in self.spans:
            location_hint = f"  --> {span.file}:{span.start.line}"
            snippet = [location_hint]

            # all the lines of code in the file this comes from
            source: List[str] = SourceTracker.source(span.source_hash)

            # First, print the span from `context_start` to `start`
            # Next, sprint the focus of the span from `start` to `end`
            # If the actual span is only 1 line long, use `column` information to highlight the exact problem
            # Finally, print end context from `end` to `context_end`
            if span.context_start:
                snippet += self._format_code_segment(
                    span.context_start, span.start.previous_line(), source, span
                )
            snippet += self._format_code_segment(span.start, span.end, source, span)
            # Currently, only span highlighting if it's a one line span
            if span.start.line == span.end.line:
                error = with_color(Fore.RED, (span.end.col - span.start.col) * "^")
                snippet.append(
                    self._format_line_number(span, None)
                    + " " * (span.start.col - 1)
                    + error
                )
            if span.context_end:
                snippet += self._format_code_segment(
                    span.end.next_line(), span.context_end, source, span
                )

            snippets.append("\n".join(snippet))
        snippet_str = "\n".join(snippets)
        if self.help:
            help_str = f"= {with_color(Fore.CYAN, 'help', bold=True)}: {self.help}"
        else:
            help_str = ""
        return f"{header}\n{snippet_str}\n{help_str}\n{with_color(Fore.RED, self.long_msg or '')}\n"


@attr.s(frozen=True, eq=True)
class InvalidPatternError(ErrorWithSpan):
    code = INVALID_PATTERN_EXIT_CODE
    level = Level.ERROR


@attr.s(frozen=True, eq=True)
class InvalidRuleSchemaError(ErrorWithSpan):
    code = INVALID_PATTERN_EXIT_CODE
    level = Level.ERROR


@attr.s(frozen=True, eq=True)
class UnknownLanguageError(ErrorWithSpan):
    code = INVALID_LANGUAGE_EXIT_CODE
    level = Level.ERROR


@attr.s(frozen=True, eq=True)
class InvalidPatternNameError(InvalidRuleSchemaError):
    pass


@attr.s(frozen=True, eq=True)
class SourceParseError(ErrorWithSpan):
    code = INVALID_CODE_EXIT_CODE
    level = Level.WARN


@attr.s(frozen=True, eq=True)
class MatchTimeoutError(SemgrepError):
    path: Path = attr.ib()
    rule_id: str = attr.ib()

    code = MATCH_TIMEOUT_EXIT_CODE
    level = Level.WARN

    def __str__(self) -> str:
        msg = f"Warning: Semgrep exceeded timeout when running {self.rule_id} on {self.path}. See `--timeout` for more info."
        return with_color(Fore.RED, msg)

    def to_dict_base(self) -> Dict[str, Any]:
        return {
            "path": str(self.path),
            "rule_id": self.rule_id,
        }


@attr.s(frozen=True, eq=True)
class TooManyMatchesError(SemgrepError):
    path: Path = attr.ib()
    rule_id: str = attr.ib()

    code = TOO_MANY_MATCHES_EXIT_CODE
    level = Level.WARN

    def __str__(self) -> str:
        msg = f"Warning: Semgrep exceeded number of matches when running {self.rule_id} on {self.path}."
        return with_color(Fore.RED, msg)

    def to_dict_base(self) -> Dict[str, Any]:
        return {
            "path": str(self.path),
            "rule_id": self.rule_id,
        }


@attr.s(frozen=True, eq=True)
class OutOfMemoryError(SemgrepError):
    path: Path = attr.ib()
    rule_id: str = attr.ib()

    code = MATCH_MAX_MEMORY_EXIT_CODE
    level = Level.WARN

    def __str__(self) -> str:
        msg = f"Warning: Semgrep exceeded memory when running {self.rule_id} on {self.path}. See `--max-memory` for more info."
        return with_color(Fore.RED, msg)

    def to_dict_base(self) -> Dict[str, Any]:
        return {
            "path": str(self.path),
            "rule_id": self.rule_id,
        }


@attr.s(frozen=True, eq=True)
class LexicalError(SemgrepError):
    path: Path = attr.ib()
    rule_id: str = attr.ib()

    code = LEXICAL_ERROR_EXIT_CODE
    level = Level.WARN

    def __str__(self) -> str:
        msg = f"Warning: Semgrep encountered a lexical error when running {self.rule_id} on {self.path}. Please ensure this is valid code."
        return with_color(Fore.RED, msg)

    def to_dict_base(self) -> Dict[str, Any]:
        return {
            "path": str(self.path),
            "rule_id": self.rule_id,
        }


class _UnknownLanguageError(SemgrepInternalError):
    pass


class _UnknownExtensionError(SemgrepInternalError):
    pass



def test_severity_error(run_semgrep_in_tmp, snapshot):
    json_str = run_semgrep_in_tmp("rules/inside.yaml", options=["--severity", "ERROR"])
    assert json_str != ""
    assert '"severity": "INFO"' not in json_str
    assert '"severity": "WARNING"' not in json_str


def test_severity_info(run_semgrep_in_tmp, snapshot):
    # Shouldn't return errors or results, since inside.yaml has 'severity: ERROR'
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/inside.yaml", options=["--severity", "INFO"]),
        "results.json",
    )


def test_severity_warning(run_semgrep_in_tmp, snapshot):
    # Shouldn't return errors or results, since inside.yaml has 'severity: ERROR'
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/inside.yaml", options=["--severity", "WARNING"]),
        "results.json",
    )


def test_severity_multiple(run_semgrep_in_tmp, snapshot):
    # Shouldn't return errors or results, since inside.yaml has 'severity: ERROR'
    # Differs from the two preceding tests in that we're testing adding multiple severity strings
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/inside.yaml", options=["--severity", "INFO", "--severity", "WARNING"]
        ),
        "results.json",
    )



from typing import Dict
from typing import List
from typing import Tuple

from semgrep.error import _UnknownExtensionError
from semgrep.error import _UnknownLanguageError
from semgrep.semgrep_types import FileExtension
from semgrep.semgrep_types import Language


# coupling: if you add a constant here, modify also ALL_EXTENSIONS below
# and you probably also need to update _LANGS_TO_EXTS_INTERNAL
# You may also have to regenerate some test snapshots with
# pipenv run pytest tests/e2e/test_rule_parser.py --snapshot-update
PYTHON_EXTENSIONS = [FileExtension(".py"), FileExtension(".pyi")]
JAVASCRIPT_EXTENSIONS = [FileExtension(".js"), FileExtension(".jsx")]
TYPESCRIPT_EXTENSIONS = [FileExtension(".ts"), FileExtension(".tsx")]
JAVA_EXTENSIONS = [FileExtension(".java")]
C_EXTENSIONS = [FileExtension(".c")]
GO_EXTENSIONS = [FileExtension(".go")]
RUBY_EXTENSIONS = [FileExtension(".rb")]
PHP_EXTENSIONS = [FileExtension(".php")]
LUA_EXTENSIONS = [FileExtension(".lua")]
CSHARP_EXTENSIONS = [FileExtension(".cs")]
RUST_EXTENSIONS = [FileExtension(".rs")]
KOTLIN_EXTENSIONS = [FileExtension(".kt"), FileExtension(".kts"), FileExtension(".ktm")]
YAML_EXTENSIONS = [FileExtension(".yaml"), FileExtension(".yml")]
ML_EXTENSIONS = [FileExtension(".mli"), FileExtension(".ml")]
JSON_EXTENSIONS = [FileExtension(".json")]

# This is used to determine the set of files with known extensions,
# i.e. those for which we have a proper parser.
ALL_EXTENSIONS = (
    PYTHON_EXTENSIONS
    + JAVASCRIPT_EXTENSIONS
    + TYPESCRIPT_EXTENSIONS
    + JAVA_EXTENSIONS
    + C_EXTENSIONS
    + GO_EXTENSIONS
    + RUBY_EXTENSIONS
    + ML_EXTENSIONS
    + JSON_EXTENSIONS
    + RUST_EXTENSIONS
    + KOTLIN_EXTENSIONS
    + YAML_EXTENSIONS
)

# This is used to select the files suitable for spacegrep, which is
# all of them. It is spacegrep itself that will detect and ignore binary
# files.
GENERIC_EXTENSIONS = [FileExtension("")]

PYTHON_LANGUAGES = [
    Language("python"),
    Language("python2"),
    Language("python3"),
    Language("py"),
]
JAVASCRIPT_LANGUAGES = [Language("javascript"), Language("js")]
TYPESCRIPT_LANGUAGES = [Language("typescript"), Language("ts")]
JAVA_LANGUAGES = [Language("java")]
C_LANGUAGES = [Language("c")]
GO_LANGUAGES = [Language("golang"), Language("go")]
RUBY_LANGUAGES = [Language("ruby"), Language("rb")]
PHP_LANGUAGES = [Language("php")]
LUA_LANGUAGES = [Language("lua")]
CSHARP_LANGUAGES = [Language("csharp"), Language("cs"), Language("C#")]
RUST_LANGUAGES = [Language("rust"), Language("Rust"), Language("rs")]
KOTLIN_LANGUAGES = [Language("kotlin"), Language("Kotlin"), Language("kt")]
YAML_LANGUAGES = [Language("yaml"), Language("Yaml")]
ML_LANGUAGES = [Language("ocaml"), Language("ml")]
JSON_LANGUAGES = [Language("json"), Language("JSON"), Language("Json")]
REGEX_LANGUAGES = [Language("none"), Language("regex")]
GENERIC_LANGUAGES = [Language("generic")]


# don't use this directly because it won't be O(1) lookup
# it's just for human readability, we process it below
_LANGS_TO_EXTS_INTERNAL: List[Tuple[List[Language], List[FileExtension]]] = [
    (PYTHON_LANGUAGES, PYTHON_EXTENSIONS),
    (JAVASCRIPT_LANGUAGES, JAVASCRIPT_EXTENSIONS),
    (TYPESCRIPT_LANGUAGES, TYPESCRIPT_EXTENSIONS),
    (JAVA_LANGUAGES, JAVA_EXTENSIONS),
    (C_LANGUAGES, C_EXTENSIONS),
    (GO_LANGUAGES, GO_EXTENSIONS),
    (ML_LANGUAGES, ML_EXTENSIONS),
    (RUBY_LANGUAGES, RUBY_EXTENSIONS),
    (PHP_LANGUAGES, PHP_EXTENSIONS),
    (JSON_LANGUAGES, JSON_EXTENSIONS),
    (LUA_LANGUAGES, LUA_EXTENSIONS),
    (CSHARP_LANGUAGES, CSHARP_EXTENSIONS),
    (RUST_LANGUAGES, RUST_EXTENSIONS),
    (KOTLIN_LANGUAGES, KOTLIN_EXTENSIONS),
    (YAML_LANGUAGES, YAML_EXTENSIONS),
    (REGEX_LANGUAGES + GENERIC_LANGUAGES, GENERIC_EXTENSIONS),
]


def all_supported_languages() -> List[Language]:
    """
    We want the list of languages to be deterministic, so sort it
    """
    return sorted(sum((languages for languages, _ in _LANGS_TO_EXTS_INTERNAL), []))


# create a dictionary for fast lookup and reverse lookup
_LANGS_TO_EXTS: Dict[Language, List[FileExtension]] = {}
_EXTS_TO_LANGS: Dict[FileExtension, List[Language]] = {}
for languages, extensions in _LANGS_TO_EXTS_INTERNAL:
    for lang in languages:
        _LANGS_TO_EXTS[lang] = extensions
    for extension in extensions:
        _EXTS_TO_LANGS[extension] = languages


def ext_to_langs(ext: FileExtension) -> List[Language]:
    langs = _EXTS_TO_LANGS.get(ext)
    if langs is None:
        raise _UnknownExtensionError(f"Unsupported extension: {ext}")
    return langs


def lang_to_exts(language: Language) -> List[FileExtension]:
    """
    Convert language to expected file extensions

    If language is not a supported semgrep language then
    raises _UnknownLanguageError
    """
    extensions = _LANGS_TO_EXTS.get(language)
    if extensions is None:
        raise _UnknownLanguageError(f"Unsupported Language: {language}")
    return extensions



from typing import Any
from typing import Dict
from typing import List

from semgrep.semgrep_types import Language


class Equivalence:
    """
    Equivalence subunit.
    YAML spec. See https://github.com/returntocorp/semgrep/pull/362#issuecomment-607249377
    '''
    equivalences:
      - id: plus-commutative
        pattern: $X + $Y <==> $Y + $X
        languages: [python, javascript, c, go, java]
      - id: eq-to-no-eq
        pattern: $X == $X ==> $X != $X
        languages: [python, javascript, c, go, java]
    '''
    """

    def __init__(self, equiv_id: str, pattern: str, languages: List[Language]):
        self._id = equiv_id
        self._pattern = pattern
        self._languages = languages

    @property
    def pattern(self) -> str:
        return self._pattern

    def to_json(self) -> Dict[str, Any]:
        return {"id": self._id, "pattern": self._pattern, "languages": self._languages}



import contextlib
import json
import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import List
from typing import Optional
from typing import Union

import appdirs
import pytest

from . import public_repos
from semgrep.constants import OutputFormat

TESTS_PATH = Path(__file__).parent

MASKED_KEYS = [
    "tool.driver.semanticVersion",
    "results.extra.metavars.*.unique_id.md5sum",
    "results.*.checks.*.matches",
]


def mark_masked(obj, path):
    _mark_masked(obj, path.split("."))


def _mark_masked(obj, path_items):
    key = path_items[0]
    if len(path_items) == 1 and key in obj:
        obj[key] = "<masked in tests>"
    else:
        if key == "*":
            next_obj = list(obj.values())
        else:
            next_obj = obj.get(key)
        if next_obj is None:
            next_objs = []
        elif not isinstance(next_obj, list):
            next_objs = [next_obj]
        else:
            next_objs = next_obj
        for o in next_objs:
            if isinstance(o, dict):
                _mark_masked(o, path_items[1:])


def _clean_output_json(output_json: str) -> str:
    """Make semgrep's output deterministic and nicer to read."""
    output = json.loads(output_json)
    for path in MASKED_KEYS:
        mark_masked(output, path)

    return json.dumps(output, indent=2, sort_keys=True)


def _run_semgrep(
    config: Optional[Union[str, Path, List[str]]] = None,
    *,
    target_name: str = "basic",
    options: Optional[List[Union[str, Path]]] = None,
    output_format: OutputFormat = OutputFormat.JSON,
    stderr: bool = False,
    strict: bool = True,
) -> str:
    """Run the semgrep CLI.

    :param config: what to pass as --config's value
    :param target_name: which directory within ./e2e/targets/ to scan
    :param options: additional CLI flags to add
    :param output_format: which format to use
    :param stderr: whether to merge stderr into the returned string
    """
    if options is None:
        options = []

    if strict:
        options.append("--strict")

    options.append("--disable-version-check")

    if config is not None:
        if isinstance(config, list):
            for conf in config:
                options.extend(["--config", conf])
        else:
            options.extend(["--config", config])

    if output_format == OutputFormat.JSON:
        options.append("--json")
    elif output_format == OutputFormat.JUNIT_XML:
        options.append("--junit-xml")
    elif output_format == OutputFormat.SARIF:
        options.append("--sarif")

    output = subprocess.check_output(
        [sys.executable, "-m", "semgrep", *options, Path("targets") / target_name],
        encoding="utf-8",
        stderr=subprocess.STDOUT if stderr else subprocess.PIPE,
    )

    if output_format.is_json() and not stderr:
        output = _clean_output_json(output)

    return output


REPO_CACHE = Path(
    os.path.expanduser(
        os.environ.get("GITHUB_REPO_CACHE", appdirs.user_cache_dir("semgrep-tests"))
    )
)


@pytest.fixture()
def clone_github_repo():
    """
    Fixture to clone a github repo. Usage:
    ```
    def my_test_function(clone_github_repo):
        repo_path = clone_github_repo(url="https://github.com/returntocorp/semgrep", sha="abdfe')
        subprocess.run(["ls", repo_path])
    ```

    :returns: A path to the repo, guaranteed to live at least until the end of the test
    """
    yield _github_repo_retry_wrapper


@contextlib.contextmanager
def chdir(dirname=None):
    curdir = os.getcwd()
    try:
        if dirname is not None:
            os.chdir(dirname)
        yield
    finally:
        os.chdir(curdir)


def _github_repo_retry_wrapper(
    repo_url: str, sha: Optional[str] = None, retries: int = 3
):
    """
    Internal fixture function. Do not use directly, use the `clone_github_repo` fixture.
    Wraps `_github_repo` function with retries. If the `_github_repo` throws an exception,
    it will delete `repo_destination` and retry up to `retries` times.
    """
    sha_str = sha or "latest"
    repo_dir = "-".join(repo_url.split("/")[-2:]) + "-" + sha_str
    repo_destination = REPO_CACHE / repo_dir
    try:
        return _github_repo(repo_url, sha, repo_destination)
    except (GitError, subprocess.CalledProcessError) as ex:
        print(f"Failed to clone github repo for tests {ex}")
        if repo_destination.exists():
            shutil.rmtree(repo_destination)
        if retries == 0:
            raise
        else:
            return _github_repo_retry_wrapper(repo_url, sha, retries - 1)


class GitError(BaseException):
    pass


def _github_repo(repo_url: str, sha: Optional[str], repo_destination: Path):
    """
    Internal fixture function. Use the `clone_github_repo` fixture.
    Clones the github repo at repo_url into `repo_destination` and checks out `sha`.

    If `repo_destination` already exists, it will validate that the correct repo is present at that location.
    """
    if not repo_destination.exists():
        if sha is None:
            subprocess.check_output(
                ["git", "clone", "--depth=1", repo_url, repo_destination]
            )
        else:
            repo_destination.mkdir()
            # Sadly, no fast way to clone a specific commit without a super
            # modern git client
            subprocess.check_output(["git", "clone", repo_url, repo_destination])
            with chdir(repo_destination):
                subprocess.check_output(["git", "checkout", sha])

    # validate that the repo seems setup properly
    with chdir(repo_destination):
        # some tests modify it, lets put everything back to normal
        subprocess.check_output(["git", "clean", "-fd"])
        subprocess.check_output(["git", "reset", "--hard"])
        all_clean = (
            subprocess.check_output(["git", "status", "--porcelain"]).strip() == b""
        )
        if not all_clean:
            raise GitError("Couldn't clean the repo, something is wrong. Deleting.")
        repo_sha = subprocess.check_output(["git", "rev-parse", "HEAD"])
        if sha:
            if not repo_sha.startswith(sha.encode("utf-8")):
                shutil.rmtree(repo_destination)
                raise GitError(
                    f"Github repo is broken (not set to correct sha: {repo_sha.decode('utf-8', errors='replace')}"
                )

    return repo_destination


@pytest.fixture
def run_semgrep_in_tmp(monkeypatch, tmp_path):
    (tmp_path / "targets").symlink_to(Path(TESTS_PATH / "e2e" / "targets").resolve())
    (tmp_path / "rules").symlink_to(Path(TESTS_PATH / "e2e" / "rules").resolve())

    monkeypatch.chdir(tmp_path)

    yield _run_semgrep


def pytest_addoption(parser):
    parser.addoption(
        "--qa",
        action="store_true",
        dest="is_qa",
        default=False,
        help="enable comprehensive QA tests",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "qa: mark tests that only need to run during QA")


def pytest_runtest_setup(item):
    if item.get_closest_marker("qa") and not item.config.getoption("--qa"):
        pytest.skip("skipping QA tests, add --qa flag to run them")


# Add `public_repo_url` as a parameter to your test to use this. It generates 1 test case per repo url in `ALL_REPOS`
def pytest_generate_tests(metafunc):
    if "public_repo_url" in metafunc.fixturenames:
        repos = public_repos.ALL_REPOS
        metafunc.parametrize("public_repo_url", repos, ids=lambda r: r["repo"])



import logging
import os
import time
from collections import OrderedDict
from pathlib import Path
from typing import Any
from typing import Dict
from typing import IO
from typing import List
from typing import Optional
from typing import Tuple

from ruamel.yaml import YAML
from ruamel.yaml import YAMLError

from semgrep.constants import CLI_RULE_ID
from semgrep.constants import DEFAULT_CONFIG_FILE
from semgrep.constants import DEFAULT_CONFIG_FOLDER
from semgrep.constants import DEFAULT_SEMGREP_CONFIG_NAME
from semgrep.constants import ID_KEY
from semgrep.constants import PLEASE_FILE_ISSUE_TEXT
from semgrep.constants import RULES_KEY
from semgrep.constants import SEMGREP_URL
from semgrep.constants import SEMGREP_USER_AGENT
from semgrep.error import InvalidRuleSchemaError
from semgrep.error import SemgrepError
from semgrep.error import UNPARSEABLE_YAML_EXIT_CODE
from semgrep.rule import Rule
from semgrep.rule_lang import parse_yaml_preserve_spans
from semgrep.rule_lang import Span
from semgrep.rule_lang import YamlMap
from semgrep.rule_lang import YamlTree
from semgrep.util import is_config_suffix
from semgrep.util import is_url

logger = logging.getLogger(__name__)

IN_DOCKER = "SEMGREP_IN_DOCKER" in os.environ
IN_GH_ACTION = "GITHUB_WORKSPACE" in os.environ

SRC_DIRECTORY = Path(os.environ.get("SEMGREP_SRC_DIRECTORY", Path("/") / "src"))
OLD_SRC_DIRECTORY = Path("/") / "home" / "repo"

RULES_REGISTRY = {"r2c": "https://semgrep.dev/c/p/r2c"}

MISSING_RULE_ID = "no-rule-id"

DEFAULT_CONFIG = {
    "rules": [
        {
            "id": "eqeq-is-bad",
            "pattern": "$X == $X",
            "message": "$X == $X is a useless equality check",
            "languages": ["python"],
            "severity": "ERROR",
        },
    ],
}


class Config:
    def __init__(self, valid_configs: Dict[str, List[Rule]]) -> None:
        """
        Handles parsing and validating of config files
        and exposes ability to get all rules in parsed config files
        """
        self.valid = valid_configs

    @classmethod
    def from_pattern_lang(
        cls, pattern: str, lang: str
    ) -> Tuple["Config", List[SemgrepError]]:
        config_dict = manual_config(pattern, lang)
        valid, errors = cls._validate(config_dict)
        return cls(valid), errors

    @classmethod
    def from_config_list(
        cls, configs: List[str]
    ) -> Tuple["Config", List[SemgrepError]]:
        """
        Takes in list of files/directories and returns Config object as well as
        list of errors parsing said config files

        If empty list is passed, tries to read config file at default locations
        """
        config_dict: Dict[str, YamlTree] = {}
        errors: List[SemgrepError] = []

        if not configs:
            try:
                config_dict.update(load_default_config())
            except SemgrepError as e:
                errors.append(e)

        for i, config in enumerate(configs):
            try:
                # Patch config_id to fix https://github.com/returntocorp/semgrep/issues/1912
                resolved_config = resolve_config(config)
                if not resolved_config:
                    logger.debug(f"Could not resolve config for {config}. Skipping.")
                    continue

                for (
                    resolved_config_key,
                    resolved_config_yaml_tree,
                ) in resolved_config.items():
                    patched_resolved_config: Dict[str, YamlTree] = {}
                    patched_resolved_config[
                        f"{resolved_config_key}_{i}"
                    ] = resolved_config_yaml_tree

                    config_dict.update(patched_resolved_config)
            except SemgrepError as e:
                errors.append(e)

        valid, parse_errors = cls._validate(config_dict)
        errors.extend(parse_errors)
        return cls(valid), errors

    def get_rules(self, no_rewrite_rule_ids: bool) -> List[Rule]:
        """
        Return list of rules

        If no_rewrite_rule_ids is True will not add
        path to config file to start of rule_ids
        """
        configs = self.valid
        if not no_rewrite_rule_ids:
            # re-write the configs to have the hierarchical rule ids
            configs = self._rename_rule_ids(configs)

        return list(
            OrderedDict.fromkeys([rule for rules in configs.values() for rule in rules])
        )

    @staticmethod
    def _safe_relative_to(a: Path, b: Path) -> Path:
        try:
            return a.relative_to(b)
        except ValueError:
            # paths had no common prefix; not possible to relativize
            return a

    @staticmethod
    def _convert_config_id_to_prefix(config_id: str) -> str:
        at_path = Path(config_id)
        at_path = Config._safe_relative_to(at_path, Path.cwd())

        prefix = ".".join(at_path.parts[:-1]).lstrip("./").lstrip(".")
        if len(prefix):
            prefix += "."
        return prefix

    @staticmethod
    def _rename_rule_ids(valid_configs: Dict[str, List[Rule]]) -> Dict[str, List[Rule]]:
        transformed = {}
        for config_id, rules in valid_configs.items():
            transformed[config_id] = [
                rule.with_id(
                    f"{Config._convert_config_id_to_prefix(config_id)}{rule.id or MISSING_RULE_ID}"
                )
                for rule in rules
            ]
        return transformed

    # the mypy ignore is cause YamlTree puts an Any inside the @staticmethod decorator
    @staticmethod
    def _validate(  # type: ignore[misc]
        config_dict: Dict[str, YamlTree]
    ) -> Tuple[Dict[str, List[Rule]], List[SemgrepError]]:
        """
        Take configs and separate into valid and list of errors parsing the invalid ones
        """
        errors: List[SemgrepError] = []
        valid: Dict[str, Any] = {}
        for config_id, config_yaml_tree in config_dict.items():
            config = config_yaml_tree.value
            if not isinstance(config, YamlMap):
                errors.append(SemgrepError(f"{config_id} was not a mapping"))
                continue

            rules = config.get(RULES_KEY)
            if rules is None:
                errors.append(
                    InvalidRuleSchemaError(
                        short_msg="missing keys",
                        long_msg=f"{config_id} is missing `{RULES_KEY}` as top-level key",
                        spans=[config_yaml_tree.span.truncate(lines=5)],
                    )
                )
                continue
            valid_rules = []
            for rule_dict in rules.value:

                try:
                    rule = validate_single_rule(config_id, rule_dict)
                except InvalidRuleSchemaError as ex:
                    errors.append(ex)
                else:
                    valid_rules.append(rule)

            if valid_rules:
                valid[config_id] = valid_rules
        return valid, errors


def validate_single_rule(
    config_id: str, rule_yaml: YamlTree[YamlMap]
) -> Optional[Rule]:
    """
    Validate that a rule dictionary contains all necessary keys
    and can be correctly parsed.
    """
    rule: YamlMap = rule_yaml.value

    # Defaults to search mode if mode is not specified
    return Rule.from_yamltree(rule_yaml)


def manual_config(pattern: str, lang: str) -> Dict[str, YamlTree]:
    # TODO remove when using sgrep -e ... -l ... instead of this hacked config
    pattern_span = Span.from_string(pattern, filename="CLI Input")
    pattern_tree = YamlTree[str](value=pattern, span=pattern_span)
    error_span = Span.from_string(
        f"Semgrep bug generating manual config {PLEASE_FILE_ISSUE_TEXT}", filename=None
    )
    return {
        "manual": YamlTree.wrap(
            {
                RULES_KEY: [
                    {
                        ID_KEY: CLI_RULE_ID,
                        "pattern": pattern_tree,
                        "message": pattern,
                        "languages": [lang],
                        "severity": "ERROR",
                    }
                ]
            },
            span=error_span,
        )
    }


def resolve_targets(targets: List[str]) -> List[Path]:
    base_path = get_base_path()
    return [
        Path(target) if Path(target).is_absolute() else base_path.joinpath(target)
        for target in targets
    ]


def adjust_for_docker() -> None:
    # change into this folder so that all paths are relative to it
    if IN_DOCKER and not IN_GH_ACTION:
        if OLD_SRC_DIRECTORY.exists():
            raise SemgrepError(
                f"Detected Docker environment using old code volume, please use '{SRC_DIRECTORY}' instead of '{OLD_SRC_DIRECTORY}'"
            )
        if not SRC_DIRECTORY.exists():
            raise SemgrepError(
                f"Detected Docker environment without a code volume, please include '-v \"${{PWD}}:{SRC_DIRECTORY}\"'"
            )
    if SRC_DIRECTORY.exists():
        os.chdir(SRC_DIRECTORY)


def get_base_path() -> Path:
    return Path(os.curdir)


def indent(msg: str) -> str:
    return "\n".join(["\t" + line for line in msg.splitlines()])


def parse_config_at_path(
    loc: Path, base_path: Optional[Path] = None
) -> Dict[str, YamlTree]:
    """
    Assumes file at loc exists
    """
    config_id = str(loc)
    if base_path:
        config_id = str(loc).replace(str(base_path), "")

    with loc.open() as f:
        return parse_config_string(config_id, f.read(), str(loc))


def parse_config_string(
    config_id: str, contents: str, filename: Optional[str]
) -> Dict[str, YamlTree]:
    if not contents:
        raise SemgrepError(
            f"Empty configuration file {filename}",
            code=UNPARSEABLE_YAML_EXIT_CODE,
        )
    try:
        data = parse_yaml_preserve_spans(contents, filename)
        return {config_id: data}
    except YAMLError as se:
        raise SemgrepError(
            f"Invalid YAML file {config_id}:\n{indent(str(se))}",
            code=UNPARSEABLE_YAML_EXIT_CODE,
        )


def parse_config_folder(loc: Path, relative: bool = False) -> Dict[str, YamlTree]:
    configs = {}
    for l in loc.rglob("*"):
        # Allow manually specified paths with ".", but don't auto-expand them
        correct_suffix = is_config_suffix(l)
        if not _is_hidden_config(l.relative_to(loc)) and correct_suffix:
            if l.is_file():
                configs.update(parse_config_at_path(l, loc if relative else None))
    return configs


def _is_hidden_config(loc: Path) -> bool:
    """
    Want to keep rules/.semgrep.yml but not path/.github/foo.yml
    Also want to keep src/.semgrep/bad_pattern.yml but not ./.pre-commit-config.yaml
    """
    return any(
        part != os.curdir
        and part != os.pardir
        and part.startswith(".")
        and DEFAULT_SEMGREP_CONFIG_NAME not in part
        for part in loc.parts
    )


def load_default_config() -> Dict[str, YamlTree]:
    """
    Load config from DEFAULT_CONFIG_FILE or DEFAULT_CONFIG_FOLDER
    """
    base_path = get_base_path()
    default_file = base_path.joinpath(DEFAULT_CONFIG_FILE)
    default_folder = base_path.joinpath(DEFAULT_CONFIG_FOLDER)
    if default_file.exists():
        return parse_config_at_path(default_file)
    elif default_folder.exists():
        return parse_config_folder(default_folder, relative=True)
    else:
        return {}


def load_config_from_local_path(location: str) -> Dict[str, YamlTree]:
    """
    Return config file(s) as dictionary object
    """
    base_path = get_base_path()
    loc = base_path.joinpath(location)
    if loc.exists():
        if loc.is_file():
            return parse_config_at_path(loc)
        elif loc.is_dir():
            return parse_config_folder(loc)
        else:
            raise SemgrepError(f"config location `{loc}` is not a file or folder!")
    else:
        addendum = ""
        if IN_DOCKER:
            addendum = " (since you are running in docker, you cannot specify arbitrary paths on the host; they must be mounted into the container)"
        raise SemgrepError(
            f"unable to find a config; path `{loc}` does not exist{addendum}"
        )


def nice_semgrep_url(url: str) -> str:
    """
    Alters semgrep.dev urls to let user
    click through to the nice display page instead
    of raw YAML.
    Replaces '/c/' in semgrep urls with '/'.
    """
    import urllib

    parsed = urllib.parse.urlparse(url)
    if "semgrep.dev" in parsed.netloc and parsed.path.startswith("/c"):
        return url.replace("/c/", "/")
    return url


def download_config(config_url: str) -> Dict[str, YamlTree]:
    import requests  # here for faster startup times

    DOWNLOADING_MESSAGE = f"downloading config..."
    logger.debug(f"trying to download from {config_url}")
    logger.info(
        f"using config from {nice_semgrep_url(config_url)}. Visit https://semgrep.dev/registry to see all public rules."
    )
    logger.info(DOWNLOADING_MESSAGE)
    headers = {"User-Agent": SEMGREP_USER_AGENT}

    try:
        r = requests.get(config_url, stream=True, headers=headers, timeout=20)
        if r.status_code == requests.codes.ok:
            content_type = r.headers.get("Content-Type")
            yaml_types = [
                "text/plain",
                "application/x-yaml",
                "text/x-yaml",
                "text/yaml",
                "text/vnd.yaml",
            ]
            if content_type and any((ct in content_type for ct in yaml_types)):
                return parse_config_string(
                    "remote-url",
                    r.content.decode("utf-8", errors="replace"),
                    filename=f"{config_url[:20]}...",
                )
            else:
                raise SemgrepError(
                    f"unknown content-type: {content_type} returned by config url: {config_url}. Can not parse"
                )
        else:
            raise SemgrepError(
                f"bad status code: {r.status_code} returned by config url: {config_url}"
            )
    except Exception as e:
        raise SemgrepError(f"Failed to download config from {config_url}: {str(e)}")

    return None


def is_registry_id(config_str: str) -> bool:
    """
    Starts with r/, p/, s/ for registry, pack, and snippet respectively
    """
    return config_str[:2] in {"r/", "p/", "s/"}


def is_saved_snippet(config_str: str) -> bool:
    """
    config_str is saved snippet which has format username:snippetname
    """
    return len(config_str.split(":")) == 2


def registry_id_to_url(registry_id: str) -> str:
    """
    Convert from registry_id to semgrep.dev url
    """
    return f"{SEMGREP_URL}{registry_id}"


def saved_snippet_to_url(snippet_id: str) -> str:
    """
    Convert from username:snippetname to semgrep.dev url
    """
    return registry_id_to_url(f"s/{snippet_id}")


def resolve_config(config_str: str) -> Dict[str, YamlTree]:
    """ resolves if config arg is a registry entry, a url, or a file, folder, or loads from defaults if None"""
    start_t = time.time()
    if config_str in RULES_REGISTRY:
        config = download_config(RULES_REGISTRY[config_str])
    elif is_url(config_str):
        config = download_config(config_str)
    elif is_registry_id(config_str):
        config = download_config(registry_id_to_url(config_str))
    elif is_saved_snippet(config_str):
        config = download_config(saved_snippet_to_url(config_str))
    else:
        config = load_config_from_local_path(config_str)
    if config:
        logger.debug(f"loaded {len(config)} configs in {time.time() - start_t}")
    return config


def generate_config(fd: IO, lang: Optional[str], pattern: Optional[str]) -> None:
    config = DEFAULT_CONFIG

    if lang:
        config["rules"][0]["languages"] = [lang]
    if pattern:
        config["rules"][0]["pattern"] = pattern

    yaml = YAML()

    try:
        yaml.dump(config, fd)
        logger.info(f"Template config successfully written to {fd.name}")
    except Exception as e:
        raise SemgrepError(str(e))


def get_config(
    pattern: str, lang: str, config_strs: List[str]
) -> Tuple[Config, List[SemgrepError]]:
    if pattern:
        if not lang:
            raise SemgrepError("language must be specified when a pattern is passed")
        config, errors = Config.from_pattern_lang(pattern, lang)
    else:
        config, errors = Config.from_config_list(config_strs)

    if not config:
        raise SemgrepError(
            f"No config given and {DEFAULT_CONFIG_FILE} was not found. Try running with --help to debug or if you want to download a default config, try running with --config r2c"
        )

    return config, errors



import pytest


def idfn(options):
    return "-and-".join(flag.strip("-") for flag in options if flag.startswith("--"))


@pytest.mark.parametrize(
    "options",
    [
        ["--exclude", "excluded.*"],
        ["--include", "included.*"],
        ["--exclude", "excluded"],
        ["--include", "included"],
        ["--include", "included", "--exclude", "excluded.*"],
        ["--exclude", "excluded", "--include", "included.*"],
        ["--exclude", "excluded.*", "--exclude", "included.*"],
        ["--exclude", "excluded", "--exclude", "included"],
        ["--include", "excluded.*", "--include", "included.*"],
        ["--include", "excluded", "--include", "included"],
    ],
    ids=idfn,
)
def test_exclude_include(run_semgrep_in_tmp, snapshot, options):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/eqeq.yaml",
            options=options,
            target_name="exclude_include",
        ),
        "results.json",
    )



import tempfile
from pathlib import Path

import pytest


def test_autofix(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/autofix/autofix.yaml", target_name="autofix/autofix.py"
        ),
        "results.json",
    )

    # Make a copy of the target file b/c autofixes are inline. We
    # don't want to modify the the actual target file, and there
    # isn't a way currently to dump the fixed file contents before
    # writing.
    # This tempfile will be deleted when the with context closes.
    with tempfile.NamedTemporaryFile(dir=Path("targets")) as tf:
        with open(Path("targets") / "autofix/autofix.py", "r") as fin:
            tf.write(fin.read().encode("utf-8"))
        tf.flush()  # Make sure file has been copied.
        tf.seek(
            0
        )  # Seek to beginning since Semgrep will be reading from it. Just in case.
        run_semgrep_in_tmp(
            "rules/autofix/autofix.yaml", target_name=tf.name, options=["--autofix"]
        )
        tf.seek(0)  # Seek to beginning again so we can read and compare to snapshot.
        snapshot.assert_match(
            tf.read().decode("utf-8", errors="replace"), "autofix-fixed"
        )


@pytest.mark.parametrize(
    "rule,target",
    [
        ("rules/autofix/csv-writer.yaml", "autofix/csv-writer.py"),
        ("rules/autofix/defaulthttpclient.yaml", "autofix/defaulthttpclient.java"),
        ("rules/autofix/flask-use-jsonify.yaml", "autofix/flask-use-jsonify.py"),
        ("rules/autofix/requests-use-timeout.yaml", "autofix/requests-use-timeout.py"),
        (
            "rules/autofix/django-none-password-default.yaml",
            "autofix/django-none-password-default.py",
        ),
    ],
)
def test_regex_autofix(run_semgrep_in_tmp, snapshot, rule, target):
    # Yes, this is fugly. I apologize. T_T
    snapshot.assert_match(
        run_semgrep_in_tmp(rule, target_name=target),
        "results.json",
    )
    # Make a copy of the target file b/c autofixes are inline. We
    # don't want to modify the the actual target file, and there
    # isn't a way currently to dump the fixed file contents before
    # writing.
    # This tempfile will be deleted when the with context closes.
    with tempfile.NamedTemporaryFile(dir=Path("targets")) as tf:
        with open(Path("targets") / target, "r") as fin:
            tf.write(fin.read().encode("utf-8"))
        tf.flush()  # Make sure file has been copied.
        tf.seek(
            0
        )  # Seek to beginning since Semgrep will be reading from it. Just in case.
        run_semgrep_in_tmp(rule, target_name=tf.name, options=["--autofix"])
        tf.seek(0)  # Seek to beginning again so we can read and compare to snapshot.
        snapshot.assert_match(
            tf.read().decode("utf-8", errors="replace"), f"{target}-fixed"
        )



import itertools
import logging
import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Any
from typing import Callable
from typing import IO
from typing import Iterable
from typing import List
from typing import Set
from typing import Tuple
from typing import TypeVar
from urllib.parse import urlparse

import pkg_resources
from colorama import Fore
from tqdm import tqdm

from semgrep.constants import YML_SUFFIXES
from semgrep.constants import YML_TEST_SUFFIXES

T = TypeVar("T")

global DEBUG
global QUIET
global FORCE_COLOR
DEBUG = False
QUIET = False
FORCE_COLOR = False


def is_url(url: str) -> bool:
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def debug_tqdm_write(msg: str, file: IO = sys.stderr) -> None:
    if DEBUG:
        tqdm.write(msg, file=file)


def flatten(L: Iterable[Iterable[Any]]) -> Iterable[Any]:
    for list in L:
        for item in list:
            yield item


def set_flags(debug_level: bool, quiet: bool, force_color: bool) -> None:
    """Set the global DEBUG and QUIET flags"""
    logger = logging.getLogger("semgrep")
    logger.handlers = []
    handler = logging.StreamHandler()
    formatter = logging.Formatter("%(message)s")
    handler.setFormatter(formatter)

    level = logging.INFO
    if debug_level:
        level = logging.DEBUG
    elif quiet:
        level = logging.ERROR

    handler.setLevel(level)
    logger.addHandler(handler)
    logger.setLevel(level)

    # TODO move to a proper logging framework
    global DEBUG
    global QUIET
    global FORCE_COLOR
    if debug_level:
        DEBUG = True
        # debug_print("DEBUG is on")
    if quiet:
        QUIET = True
        # debug_print("QUIET is on")

    if force_color:
        FORCE_COLOR = True
        # debug_print("Output will use ANSI escapes, even if output is not a TTY")


def partition(pred: Callable, iterable: Iterable) -> Tuple[List, List]:
    """E.g. partition(is_odd, range(10)) -> 1 3 5 7 9  and  0 2 4 6 8"""
    i1, i2 = itertools.tee(iterable)
    return list(filter(pred, i1)), list(itertools.filterfalse(pred, i2))


def partition_set(pred: Callable, iterable: Iterable) -> Tuple[Set, Set]:
    """E.g. partition(is_odd, range(10)) -> 1 3 5 7 9  and  0 2 4 6 8"""
    i1, i2 = itertools.tee(iterable)
    return set(filter(pred, i1)), set(itertools.filterfalse(pred, i2))


# cf. https://docs.python.org/3/library/itertools.html#itertools-recipes
def powerset(iterable: Iterable) -> Iterable[Tuple[Any, ...]]:
    """powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
    s = list(iterable)
    return itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(len(s) + 1)
    )


def with_color(color: str, text: str, bold: bool = False) -> str:
    """
    Wrap text in color & reset
    """
    if not sys.stderr.isatty() and not FORCE_COLOR:
        return text

    reset = Fore.RESET
    if bold:
        color = color + "\033[1m"
        reset += "\033[0m"
    return f"{color}{text}{reset}"


def progress_bar(
    iterable: Iterable[T], file: IO = sys.stderr, **kwargs: Any
) -> Iterable[T]:
    """
    Return tqdm-wrapped iterable if output stream is a tty;
    else return iterable without tqdm.
    """
    # Conditions:
    # file.isatty() - only show bar if this is an interactive terminal
    # len(iterable) > 1 - don't show progress bar when using -e on command-line. This
    #   is a hack, so it will only show the progress bar if there is more than 1 rule to run.
    # not DEBUG - don't show progress bar with debug
    # not QUIET - don't show progress bar with quiet
    listified = list(
        iterable
    )  # Consume iterable once so we can check length and then use in tqdm.
    if file.isatty() and len(listified) > 1 and not DEBUG and not QUIET:
        # mypy doesn't seem to want to follow tqdm imports. Do this to placate.
        wrapped: Iterable[T] = tqdm(listified, file=file, **kwargs)
        return wrapped
    return listified


def sub_run(cmd: List[str], **kwargs: Any) -> Any:
    """A simple proxy function to minimize and centralize subprocess usage."""
    # fmt: off
    result = subprocess.run(cmd, **kwargs)  # nosem: python.lang.security.audit.dangerous-subprocess-use.dangerous-subprocess-use
    # fmt: on
    return result


def sub_check_output(cmd: List[str], **kwargs: Any) -> Any:
    """A simple proxy function to minimize and centralize subprocess usage."""
    # fmt: off
    if QUIET:
        kwargs = {**kwargs, "stderr": subprocess.DEVNULL}
    result = subprocess.check_output(cmd, **kwargs)  # nosem: python.lang.security.audit.dangerous-subprocess-use.dangerous-subprocess-use
    # fmt: on
    return result


def compute_executable_path(exec_name: str) -> str:
    """Determine full executable path if full path is needed to run it."""
    # First, try packaged binaries
    pkg_exec = pkg_resources.resource_filename("semgrep.bin", exec_name)
    if os.path.isfile(pkg_exec):
        return pkg_exec

    # Second, try system binaries
    which_exec = shutil.which(exec_name)
    if which_exec is not None:
        return which_exec

    # Third, look for something in the same dir as the Python interpreter
    relative_path = os.path.join(os.path.dirname(sys.executable), exec_name)
    if os.path.isfile(relative_path):
        return relative_path

    raise Exception(f"Could not locate '{exec_name}' binary")


def compute_semgrep_path() -> str:
    return compute_executable_path("semgrep-core")


def compute_spacegrep_path() -> str:
    return compute_executable_path("spacegrep")


SEMGREP_PATH = compute_semgrep_path()
SPACEGREP_PATH = compute_spacegrep_path()


def liststartswith(l: List[T], head: List[T]) -> bool:
    """
    E.g.
        - liststartswith([1, 2, 3, 4], [1, 2]) -> True
        - liststartswith([1, 2, 3, 4], [1, 4]) -> False
    """
    if len(head) > len(l):
        return False

    return all(l[i] == head[i] for i in range(len(head)))


def listendswith(l: List[T], tail: List[T]) -> bool:
    """
    E.g.
        - listendswith([1, 2, 3, 4], [3, 4]) -> True
        - listendswith([1, 2, 3, 4], [1, 4]) -> False
    """
    if len(tail) > len(l):
        return False

    return all(l[len(l) - len(tail) + i] == tail[i] for i in range(len(tail)))


def is_config_suffix(path: Path) -> bool:
    return any(
        listendswith(path.suffixes, suffixes) for suffixes in YML_SUFFIXES
    ) and not is_config_test_suffix(path)


def is_config_test_suffix(path: Path) -> bool:
    return any(listendswith(path.suffixes, suffixes) for suffixes in YML_TEST_SUFFIXES)



# type: ignore
import collections
import re
import sys
import xml.dom.minidom  # nosem: python.lang.security.use-defused-xml.use-defused-xml
import xml.etree.ElementTree as ET  # nosem: python.lang.security.use-defused-xml.use-defused-xml

"""
This file is copied from: https://github.com/kyrus/python-junit-xml/blob/master/junit_xml/__init__.py

The above repository is no longer maintained, and contains minimal functionality
that mostly resides in this single file. Instead of forking and/or relying on this
unmaintained repository we (r2c) will instead pull in this single file and use what
we need.

This file is licensed under the MIT license: https://github.com/kyrus/python-junit-xml/blob/master/LICENSE.txt

Based on the understanding of what Jenkins can parse for JUnit XML files.

<?xml version="1.0" encoding="utf-8"?>
<testsuites errors="1" failures="1" tests="4" time="45">
    <testsuite errors="1" failures="1" hostname="localhost" id="0" name="test1"
               package="testdb" tests="4" timestamp="2012-11-15T01:02:29">
        <properties>
            <property name="assert-passed" value="1"/>
        </properties>
        <testcase classname="testdb.directory" name="1-passed-test" time="10"/>
        <testcase classname="testdb.directory" name="2-failed-test" time="20">
            <failure message="Assertion FAILED: failed assert" type="failure">
                the output of the testcase
            </failure>
        </testcase>
        <testcase classname="package.directory" name="3-errord-test" time="15">
            <error message="Assertion ERROR: error assert" type="error">
                the output of the testcase
            </error>
        </testcase>
        <testcase classname="package.directory" name="3-skipped-test" time="0">
            <skipped message="SKIPPED Test" type="skipped">
                the output of the testcase
            </skipped>
        </testcase>
        <testcase classname="testdb.directory" name="3-passed-test" time="10">
            <system-out>
                I am system output
            </system-out>
            <system-err>
                I am the error output
            </system-err>
        </testcase>
    </testsuite>
</testsuites>
"""


class TestSuite:
    """
    Suite of test cases.
    """

    def __init__(
        self,
        name,
        test_cases=None,
        hostname=None,
        id=None,
        package=None,
        timestamp=None,
        properties=None,
        file=None,
        log=None,
        url=None,
        stdout=None,
        stderr=None,
    ):
        self.name = name
        if not test_cases:
            test_cases = []
        try:
            iter(test_cases)
        except TypeError:
            raise TypeError("test_cases must be a list of test cases")
        self.test_cases = test_cases
        self.timestamp = timestamp
        self.hostname = hostname
        self.id = id
        self.package = package
        self.file = file
        self.log = log
        self.url = url
        self.stdout = stdout
        self.stderr = stderr
        self.properties = properties

    def build_xml_doc(self):
        """
        Builds the XML document for the JUnit test suite.
        @return: XML document with unicode string elements
        """

        # build the test suite element
        test_suite_attributes = dict()
        if any(c.assertions for c in self.test_cases):
            test_suite_attributes["assertions"] = str(
                sum([int(c.assertions) for c in self.test_cases if c.assertions])
            )
        test_suite_attributes["disabled"] = str(
            len([c for c in self.test_cases if not c.is_enabled])
        )
        test_suite_attributes["errors"] = str(
            len([c for c in self.test_cases if c.is_error()])
        )
        test_suite_attributes["failures"] = str(
            len([c for c in self.test_cases if c.is_failure()])
        )
        test_suite_attributes["name"] = str(self.name)
        test_suite_attributes["skipped"] = str(
            len([c for c in self.test_cases if c.is_skipped()])
        )
        test_suite_attributes["tests"] = str(len(self.test_cases))
        test_suite_attributes["time"] = str(
            sum(c.elapsed_sec for c in self.test_cases if c.elapsed_sec)
        )

        if self.hostname:
            test_suite_attributes["hostname"] = str(self.hostname)
        if self.id:
            test_suite_attributes["id"] = str(self.id)
        if self.package:
            test_suite_attributes["package"] = str(self.package)
        if self.timestamp:
            test_suite_attributes["timestamp"] = str(self.timestamp)
        if self.file:
            test_suite_attributes["file"] = str(self.file)
        if self.log:
            test_suite_attributes["log"] = str(self.log)
        if self.url:
            test_suite_attributes["url"] = str(self.url)

        xml_element = ET.Element("testsuite", test_suite_attributes)

        # add any properties
        if self.properties:
            props_element = ET.SubElement(xml_element, "properties")
            for k, v in self.properties.items():
                attrs = {"name": str(k), "value": str(v)}
                ET.SubElement(props_element, "property", attrs)

        # add test suite stdout
        if self.stdout:
            stdout_element = ET.SubElement(xml_element, "system-out")
            stdout_element.text = str(self.stdout)

        # add test suite stderr
        if self.stderr:
            stderr_element = ET.SubElement(xml_element, "system-err")
            stderr_element.text = str(self.stderr)

        # test cases
        for case in self.test_cases:
            test_case_attributes = dict()
            test_case_attributes["name"] = str(case.name)
            if case.assertions:
                # Number of assertions in the test case
                test_case_attributes["assertions"] = "%d" % case.assertions
            if case.elapsed_sec:
                test_case_attributes["time"] = "%f" % case.elapsed_sec
            if case.timestamp:
                test_case_attributes["timestamp"] = str(case.timestamp)
            if case.classname:
                test_case_attributes["classname"] = str(case.classname)
            if case.status:
                test_case_attributes["status"] = str(case.status)
            if case.category:
                test_case_attributes["class"] = str(case.category)
            if case.file:
                test_case_attributes["file"] = str(case.file)
            if case.line:
                test_case_attributes["line"] = str(case.line)
            if case.log:
                test_case_attributes["log"] = str(case.log)
            if case.url:
                test_case_attributes["url"] = str(case.url)

            test_case_element = ET.SubElement(
                xml_element, "testcase", test_case_attributes
            )

            # failures
            for failure in case.failures:
                if failure["output"] or failure["message"]:
                    attrs = {"type": "failure"}
                    if failure["message"]:
                        attrs["message"] = str(failure["message"])
                    if failure["type"]:
                        attrs["type"] = str(failure["type"])
                    failure_element = ET.Element("failure", attrs)
                    if failure["output"]:
                        failure_element.text = str(failure["output"])
                    test_case_element.append(failure_element)

            # errors
            for error in case.errors:
                if error["message"] or error["output"]:
                    attrs = {"type": "error"}
                    if error["message"]:
                        attrs["message"] = str(error["message"])
                    if error["type"]:
                        attrs["type"] = str(error["type"])
                    error_element = ET.Element("error", attrs)
                    if error["output"]:
                        error_element.text = str(error["output"])
                    test_case_element.append(error_element)

            # skippeds
            for skipped in case.skipped:
                attrs = {"type": "skipped"}
                if skipped["message"]:
                    attrs["message"] = str(skipped["message"])
                skipped_element = ET.Element("skipped", attrs)
                if skipped["output"]:
                    skipped_element.text = str(skipped["output"])
                test_case_element.append(skipped_element)

            # test stdout
            if case.stdout:
                stdout_element = ET.Element("system-out")
                stdout_element.text = str(case.stdout)
                test_case_element.append(stdout_element)

            # test stderr
            if case.stderr:
                stderr_element = ET.Element("system-err")
                stderr_element.text = str(case.stderr)
                test_case_element.append(stderr_element)

        return xml_element


def to_xml_report_string(test_suites, prettyprint=True):
    """
    Returns the string representation of the JUnit XML document.
    @return: unicode string
    """

    try:
        iter(test_suites)
    except TypeError:
        raise TypeError("test_suites must be a list of test suites")

    xml_element = ET.Element("testsuites")
    attributes = collections.defaultdict(int)
    for ts in test_suites:
        ts_xml = ts.build_xml_doc()
        for key in ["disabled", "errors", "failures", "tests"]:
            attributes[key] += int(ts_xml.get(key, 0))
        for key in ["time"]:
            attributes[key] += float(ts_xml.get(key, 0))
        xml_element.append(ts_xml)
    for key, value in attributes.items():
        xml_element.set(key, str(value))

    xml_string = ET.tostring(xml_element)
    xml_string = _clean_illegal_xml_chars(xml_string.decode("utf-8", errors="replace"))

    if prettyprint:
        xml_string = xml_string.encode("utf-8")
        # fmt: off
        xml_string = xml.dom.minidom.parseString(xml_string)  # nosem: contrib.dlint.dlint-equivalent.insecure-xml-use
        # fmt: on
        xml_string = xml_string.toprettyxml()
    return xml_string


def _clean_illegal_xml_chars(string_to_clean):
    """
    Removes any illegal unicode characters from the given XML string.

    @see: http://stackoverflow.com/questions/1707890/fast-way-to-filter-illegal-xml-unicode-chars-in-python
    """

    illegal_unichrs = [
        (0x00, 0x08),
        (0x0B, 0x1F),
        (0x7F, 0x84),
        (0x86, 0x9F),
        (0xD800, 0xDFFF),
        (0xFDD0, 0xFDDF),
        (0xFFFE, 0xFFFF),
        (0x1FFFE, 0x1FFFF),
        (0x2FFFE, 0x2FFFF),
        (0x3FFFE, 0x3FFFF),
        (0x4FFFE, 0x4FFFF),
        (0x5FFFE, 0x5FFFF),
        (0x6FFFE, 0x6FFFF),
        (0x7FFFE, 0x7FFFF),
        (0x8FFFE, 0x8FFFF),
        (0x9FFFE, 0x9FFFF),
        (0xAFFFE, 0xAFFFF),
        (0xBFFFE, 0xBFFFF),
        (0xCFFFE, 0xCFFFF),
        (0xDFFFE, 0xDFFFF),
        (0xEFFFE, 0xEFFFF),
        (0xFFFFE, 0xFFFFF),
        (0x10FFFE, 0x10FFFF),
    ]

    illegal_ranges = [
        "%s-%s" % (chr(low), chr(high))
        for (low, high) in illegal_unichrs
        if low < sys.maxunicode
    ]

    illegal_xml_re = re.compile("[%s]" % "".join(illegal_ranges))
    return illegal_xml_re.sub("", string_to_clean)


class TestCase:
    """A JUnit test case with a result and possibly some stdout or stderr"""

    def __init__(
        self,
        name,
        classname=None,
        elapsed_sec=None,
        stdout=None,
        stderr=None,
        assertions=None,
        timestamp=None,
        status=None,
        category=None,
        file=None,
        line=None,
        log=None,
        url=None,
        allow_multiple_subelements=False,
    ):
        self.name = name
        self.assertions = assertions
        self.elapsed_sec = elapsed_sec
        self.timestamp = timestamp
        self.classname = classname
        self.status = status
        self.category = category
        self.file = file
        self.line = line
        self.log = log
        self.url = url
        self.stdout = stdout
        self.stderr = stderr

        self.is_enabled = True
        self.errors = []
        self.failures = []
        self.skipped = []
        self.allow_multiple_subalements = allow_multiple_subelements

    def add_error_info(self, message=None, output=None, error_type=None):
        """Adds an error message, output, or both to the test case"""
        error = {}
        error["message"] = message
        error["output"] = output
        error["type"] = error_type
        if self.allow_multiple_subalements:
            if message or output:
                self.errors.append(error)
        elif not len(self.errors):
            self.errors.append(error)
        else:
            if message:
                self.errors[0]["message"] = message
            if output:
                self.errors[0]["output"] = output
            if error_type:
                self.errors[0]["type"] = error_type

    def add_failure_info(self, message=None, output=None, failure_type=None):
        """Adds a failure message, output, or both to the test case"""
        failure = {}
        failure["message"] = message
        failure["output"] = output
        failure["type"] = failure_type
        if self.allow_multiple_subalements:
            if message or output:
                self.failures.append(failure)
        elif not len(self.failures):
            self.failures.append(failure)
        else:
            if message:
                self.failures[0]["message"] = message
            if output:
                self.failures[0]["output"] = output
            if failure_type:
                self.failures[0]["type"] = failure_type

    def add_skipped_info(self, message=None, output=None):
        """Adds a skipped message, output, or both to the test case"""
        skipped = {}
        skipped["message"] = message
        skipped["output"] = output
        if self.allow_multiple_subalements:
            if message or output:
                self.skipped.append(skipped)
        elif not len(self.skipped):
            self.skipped.append(skipped)
        else:
            if message:
                self.skipped[0]["message"] = message
            if output:
                self.skipped[0]["output"] = output

    def is_failure(self):
        """returns true if this test case is a failure"""
        return sum(1 for f in self.failures if f["message"] or f["output"]) > 0

    def is_error(self):
        """returns true if this test case is an error"""
        return sum(1 for e in self.errors if e["message"] or e["output"]) > 0

    def is_skipped(self):
        """returns true if this test case has been skipped"""
        return len(self.skipped) > 0



from typing import Dict
from typing import NamedTuple

Semgrep_run = NamedTuple("Semgrep_run", [("rule", str), ("target", str)])

Times = NamedTuple(
    "Times", [("parse_time", float), ("match_time", float), ("run_time", float)]
)


class ProfilingData:
    def __init__(self) -> None:
        self._match_time_matrix: Dict[Semgrep_run, Times] = {}

    def get_times(self, rule: str, target: str) -> Times:
        return self._match_time_matrix.get(
            Semgrep_run(rule=rule, target=target),
            Times(parse_time=0.0, match_time=0.0, run_time=0.0),
        )

    def set_times(self, rule: str, target: str, times: Times) -> None:
        self._match_time_matrix[Semgrep_run(rule=rule, target=target)] = times



import contextlib
import os
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Collection
from typing import Dict
from typing import Iterator
from typing import List
from typing import Set

import attr

from semgrep.config_resolver import resolve_targets
from semgrep.error import FilesNotFoundError
from semgrep.output import OutputHandler
from semgrep.semgrep_types import Language
from semgrep.target_manager_extensions import ALL_EXTENSIONS
from semgrep.target_manager_extensions import FileExtension
from semgrep.target_manager_extensions import lang_to_exts
from semgrep.util import partition_set
from semgrep.util import sub_check_output


@contextlib.contextmanager
def optional_stdin_target(target: List[str]) -> Iterator[List[str]]:
    """
    Read target input from stdin if "-" is specified
    """
    if target == ["-"]:
        try:
            with tempfile.NamedTemporaryFile(delete=False) as fd:
                fd.write(sys.stdin.buffer.read())
                fname = fd.name
            yield [fname]
        finally:
            os.remove(fname)
    else:
        yield target


@attr.s(auto_attribs=True)
class TargetManager:
    """
    Handles all file include/exclude logic for semgrep

    If respect_git_ignore is true then will only consider files that are
    tracked or (untracked but not ignored) by git

    If skip_unknown_extensions is False then targets with extensions that are
    not understood by semgrep will always be returned by get_files. Else will discard
    targets with unknown extensions
    """

    includes: List[str]
    excludes: List[str]
    targets: List[str]
    respect_git_ignore: bool
    output_handler: OutputHandler
    skip_unknown_extensions: bool

    _filtered_targets: Dict[str, Set[Path]] = attr.ib(factory=dict)

    @staticmethod
    def resolve_targets(targets: List[str]) -> Set[Path]:
        """
        Return list of Path objects appropriately resolving relative paths
        (relative to cwd) if necessary
        """
        return set(resolve_targets(targets))

    @staticmethod
    def _is_valid(path: Path) -> bool:
        return os.access(path, os.R_OK) and path.exists() and not path.is_symlink()

    @staticmethod
    def _expand_dir(
        curr_dir: Path, language: Language, respect_git_ignore: bool
    ) -> Set[Path]:
        """
        Recursively go through a directory and return list of all files with
        default file extension of language
        """

        def _parse_output(output: str, curr_dir: Path) -> Set[Path]:
            """
            Convert a newline delimited list of files to a set of path objects
            prepends curr_dir to all paths in said list

            If list is empty then returns an empty set
            """
            files: Set[Path] = set()
            if output:
                files = {
                    p
                    for p in (
                        Path(curr_dir) / elem for elem in output.strip().split("\n")
                    )
                    if TargetManager._is_valid(p)
                }
            return files

        def _find_files_with_extension(
            curr_dir: Path, extension: FileExtension
        ) -> Set[Path]:
            """
            Return set of all files in curr_dir with given extension
            """
            return {
                p
                for p in curr_dir.rglob(f"*{extension}")
                if TargetManager._is_valid(p) and p.is_file()
            }

        extensions = lang_to_exts(language)
        expanded: Set[Path] = set()

        for ext in extensions:
            if respect_git_ignore:
                try:
                    # Tracked files
                    tracked_output = sub_check_output(
                        ["git", "ls-files", f"*{ext}"],
                        cwd=curr_dir.resolve(),
                        encoding="utf-8",
                        stderr=subprocess.DEVNULL,
                    )

                    # Untracked but not ignored files
                    untracked_output = sub_check_output(
                        [
                            "git",
                            "ls-files",
                            "--other",
                            "--exclude-standard",
                            f"*{ext}",
                        ],
                        cwd=curr_dir.resolve(),
                        encoding="utf-8",
                        stderr=subprocess.DEVNULL,
                    )

                    deleted_output = sub_check_output(
                        ["git", "ls-files", "--deleted", f"*{ext}"],
                        cwd=curr_dir.resolve(),
                        encoding="utf-8",
                        stderr=subprocess.DEVNULL,
                    )
                except (subprocess.CalledProcessError, FileNotFoundError):
                    # Not a git directory or git not installed. Fallback to using rglob
                    ext_files = _find_files_with_extension(curr_dir, ext)
                    expanded = expanded.union(ext_files)
                else:
                    tracked = _parse_output(tracked_output, curr_dir)
                    untracked_unignored = _parse_output(untracked_output, curr_dir)
                    deleted = _parse_output(deleted_output, curr_dir)
                    expanded = expanded.union(tracked)
                    expanded = expanded.union(untracked_unignored)
                    expanded = expanded.difference(deleted)

            else:
                ext_files = _find_files_with_extension(curr_dir, ext)
                expanded = expanded.union(ext_files)

        return expanded

    @staticmethod
    def expand_targets(
        targets: Collection[Path], lang: Language, respect_git_ignore: bool
    ) -> Set[Path]:
        """
        Explore all directories. Remove duplicates
        """
        expanded = set()
        for target in targets:
            if not TargetManager._is_valid(target):
                continue

            if target.is_dir():
                expanded.update(
                    TargetManager._expand_dir(target, lang, respect_git_ignore)
                )
            else:
                expanded.add(target)

        return expanded

    @staticmethod
    def match_glob(path: Path, globs: List[str]) -> bool:
        """
        Return true if path or any parent of path matches any glob in globs
        """
        subpaths = [path, *path.parents]
        return any(p.match(glob) for p in subpaths for glob in globs)

    @staticmethod
    def filter_includes(arr: Set[Path], includes: List[str]) -> Set[Path]:
        """
        Returns all elements in arr that match any includes pattern

        If includes is empty, returns arr unchanged
        """
        if not includes:
            return arr

        return set(elem for elem in arr if TargetManager.match_glob(elem, includes))

    @staticmethod
    def filter_excludes(arr: Set[Path], excludes: List[str]) -> Set[Path]:
        """
        Returns all elements in arr that do not match any excludes excludes
        """
        return set(elem for elem in arr if not TargetManager.match_glob(elem, excludes))

    def filtered_files(self, lang: Language) -> Set[Path]:
        """
        Return all files that are decendants of any directory in TARGET that have
        an extension matching LANG that match any pattern in INCLUDES and do not
        match any pattern in EXCLUDES. Any file in TARGET bypasses excludes and includes.
        If a file in TARGET has a known extension that is not for langugage LANG then
        it is also filtered out

        Note also filters out any directory and decendants of `.git`
        """
        if lang in self._filtered_targets:
            return self._filtered_targets[lang]

        targets = self.resolve_targets(self.targets)

        files, directories = partition_set(lambda p: not p.is_dir(), targets)

        # Error on non-existent files
        explicit_files, nonexistent_files = partition_set(lambda p: p.is_file(), files)
        if nonexistent_files:
            self.output_handler.handle_semgrep_error(
                FilesNotFoundError(tuple(nonexistent_files))
            )

        targets = self.expand_targets(directories, lang, self.respect_git_ignore)
        targets = self.filter_includes(targets, self.includes)
        targets = self.filter_excludes(targets, self.excludes + [".git"])

        # Remove explicit_files with known extensions.
        explicit_files_with_lang_extension = set(
            f
            for f in explicit_files
            if (any(f.match(f"*{ext}") for ext in lang_to_exts(lang)))
        )
        targets = targets.union(explicit_files_with_lang_extension)

        if not self.skip_unknown_extensions:
            explicit_files_with_unknown_extensions = set(
                f
                for f in explicit_files
                if not any(f.match(f"*{ext}") for ext in ALL_EXTENSIONS)
            )
            targets = targets.union(explicit_files_with_unknown_extensions)

        self._filtered_targets[lang] = targets
        return self._filtered_targets[lang]

    def get_files(
        self, lang: Language, includes: List[str], excludes: List[str]
    ) -> List[Path]:
        """
        Returns list of files that should be analyzed for a LANG

        Given this object's TARGET, self.INCLUDE, and self.EXCLUDE will return list
        of all descendant files of directories in TARGET that end in extension
        typical for LANG. If self.INCLUDES is non empty then all files will have an ancestor
        that matches a pattern in self.INCLUDES. Will not include any file that has
        an ancestor that matches a pattern in self.EXCLUDES. Any explicitly named files
        in TARGET will bypass this global INCLUDE/EXCLUDE filter. The local INCLUDE/EXCLUDE
        filter is then applied.
        """
        targets = self.filtered_files(lang)
        targets = self.filter_includes(targets, includes)
        targets = self.filter_excludes(targets, excludes)
        return list(targets)



#!/usr/bin/env python3
import logging.handlers
import sys

from semgrep.cli import cli
from semgrep.error import OK_EXIT_CODE
from semgrep.error import SemgrepError


def main() -> int:
    # When running semgrep as a command line tool
    # silence root level logger otherwise logs higher
    # than warning are handled twice
    logger = logging.getLogger("semgrep")
    logger.propagate = False
    try:
        cli()
    # Catch custom exceptions, output the right message and exit.
    # Note: this doesn't catch all Exceptions and lets them bubble up.
    except SemgrepError as e:
        return e.code
    else:
        return OK_EXIT_CODE


if __name__ == "__main__":
    sys.exit(main())



import hashlib
import json
import re
from io import StringIO
from pathlib import Path
from typing import Any
from typing import cast
from typing import Dict
from typing import Generic
from typing import ItemsView
from typing import KeysView
from typing import List
from typing import NewType
from typing import Optional
from typing import TypeVar
from typing import Union

import attr
import jsonschema.exceptions
from jsonschema.validators import Draft7Validator
from ruamel.yaml import Node
from ruamel.yaml import RoundTripConstructor
from ruamel.yaml import YAML

from semgrep.constants import PLEASE_FILE_ISSUE_TEXT

# Do not construct SourceFileHash directly, use `SpanBuilder().add_source`
SourceFileHash = NewType("SourceFileHash", str)


class RuleSchema:
    _schema: Dict[str, Any] = {}

    @classmethod
    def get(cls) -> Dict[str, Any]:  # type: ignore
        """
        Returns the rule schema

        Not thread safe.
        """
        if not cls._schema:
            yaml = YAML()
            schema_path = Path(__file__).parent / "rule_schema.yaml"
            with schema_path.open() as fd:
                cls._schema = yaml.load(fd)
        return cls._schema


class SourceTracker:
    """
    Singleton class tracking mapping from filehashes -> file contents to support
    building error messages from Spans
    """

    # sources are a class variable to share state
    sources: Dict[SourceFileHash, List[str]] = {}

    @classmethod
    def add_source(cls, source: str) -> SourceFileHash:
        file_hash = cls._src_to_hash(source)
        cls.sources[file_hash] = source.splitlines()
        return file_hash

    @classmethod
    def source(cls, source_hash: SourceFileHash) -> List[str]:
        return cls.sources[source_hash]

    @staticmethod
    def _src_to_hash(contents: Union[str, bytes]) -> SourceFileHash:
        if isinstance(contents, str):
            contents = contents.encode("utf-8")
        return SourceFileHash(hashlib.sha256(contents).hexdigest())


@attr.s(auto_attribs=True, frozen=True, repr=False)
class Position:
    """
    Position within a file.
    :param line: 1-indexed line number
    :param col: 1-indexed column number

    line & column are 0 indexed for compatibility with semgrep-core which also produces 1-indexed results
    """

    line: int
    col: int

    def next_line(self) -> "Position":
        return attr.evolve(self, line=self.line + 1)

    def previous_line(self) -> "Position":
        return attr.evolve(self, line=self.line - 1)

    def to_dict(self) -> dict:
        return {"line": self.line, "col": self.col}

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} line={self.line} col={self.col}>"


@attr.s(auto_attribs=True, frozen=True, repr=False)
class Span:
    """
    Spans are immutable objects, representing segments of code. They have a central focus area, and
    optionally can contain surrounding context.
    """

    start: Position
    end: Position
    source_hash: SourceFileHash
    file: Optional[str]
    context_start: Optional[Position] = None
    context_end: Optional[Position] = None

    @classmethod
    def from_node(
        cls, node: Node, source_hash: SourceFileHash, filename: Optional[str]
    ) -> "Span":
        start = Position(line=node.start_mark.line + 1, col=node.start_mark.column + 1)
        end = Position(line=node.end_mark.line + 1, col=node.end_mark.column + 1)
        return Span(start=start, end=end, file=filename, source_hash=source_hash).fix()

    @classmethod
    def from_string(cls, s: str, filename: Optional[str] = None) -> "Span":
        src_hash = SourceTracker.add_source(s)
        start = Position(1, 1)
        lines = s.splitlines()
        end = Position(line=len(lines), col=len(lines[-1]))
        return Span(start=start, end=end, file=filename, source_hash=src_hash)

    def fix(self) -> "Span":
        # some issues in ruamel lead to bad spans
        # correct empty spans by rewinding to the last non-whitespace character:
        if self.start == self.end:
            src = SourceTracker.source(self.source_hash)
            cur_line = self.start.line - 1
            cur_col = self.start.col - 2
            while cur_line > 0:
                while cur_col > 0:
                    cur_char = src[cur_line][cur_col]
                    if cur_char.isspace():
                        cur_col -= 1
                    else:
                        # assign the span to the whitespace
                        start = Position(cur_line + 1, cur_col + 2)
                        end = attr.evolve(start, col=cur_col + 3)
                        return attr.evolve(self, start=start, end=end)
                cur_line -= 1
                cur_col = len(src[cur_line]) - 1
        return self

    def truncate(self, lines: int) -> "Span":
        """
        Produce a new span truncated to at most `lines` starting from the start line.
        - start_context is not considered.
        - end_context is removed
        """
        if self.end.line - self.start.line > lines:
            return attr.evolve(
                self,
                end=Position(line=self.start.line + lines, col=0),
                context_end=None,
            )
        return self

    def extend_to(self, span: "Span", context_only: bool = True) -> "Span":
        """
        Extend this span to go as far as `span`.
        :param span: The span to extend to
        :param context_only: If true, the additional lines will only be marked as context. These will be displayed,
        but unlike to core span area, they won't be highlighted in error messages.
        """
        if context_only:
            return attr.evolve(self, context_end=span.context_end or span.end)
        else:
            return attr.evolve(self, end=span.end, context_end=span.context_end)

    def with_context(
        self, before: Optional[int] = None, after: Optional[int] = None
    ) -> "Span":
        """
        Expand
        """
        new = self
        if before is not None:
            new = attr.evolve(
                new,
                context_start=Position(col=0, line=max(0, self.start.line - before)),
            )

        if after is not None:
            new = attr.evolve(
                new,
                context_end=Position(
                    col=0,
                    line=min(
                        len(SourceTracker.source(self.source_hash)),
                        self.end.line + after,
                    ),
                ),
            )
        return new

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} start={self.start} end={self.end}>"


EmptySpan = Span.from_string("a: b")

# Actually recursive but mypy is unhelpful
YamlValue = Union[str, int, List[Any], Dict[str, Any]]
LocatedYamlValue = Union[str, int, List["YamlTree"], "YamlMap"]

T = TypeVar("T", bound=LocatedYamlValue)


class YamlTree(Generic[T]):
    def __init__(self, value: T, span: Span):
        self.value = value
        self.span = span

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} span={self.span} value={self.value}>"

    def unroll_dict(self) -> Dict[str, Any]:
        """
        Helper wrapper mostly for mypy when you know it contains a dictionary
        """
        ret = self.unroll()
        if not isinstance(ret, dict):
            raise ValueError(
                f"unroll_dict called but object was actually {type(ret).__name__}"
            )
        return ret

    def unroll(self) -> YamlValue:
        """
        Recursively expand the `self.value`, converting back to a normal datastructure
        """
        if isinstance(self.value, list):
            return [x.unroll() for x in self.value]
        elif isinstance(self.value, YamlMap):
            return {str(k.unroll()): v.unroll() for k, v in self.value.items()}
        elif isinstance(self.value, YamlTree):
            return self.value.unroll()
        elif isinstance(self.value, str) or isinstance(self.value, int):
            return self.value
        else:
            raise ValueError(
                f"Invalid YAML tree structure (expected a list, dict, tree, int or str, found: {type(self.value).__name__}: {self.value}"
            )

    @classmethod
    def wrap(cls, value: YamlValue, span: Span) -> "YamlTree":  # type: ignore
        """
        Wraps a value in a YamlTree and attaches the span everywhere.
        This exists so you can take generate a datastructure from user input, but track all the errors within that
        datastructure back to the user input
        """
        if isinstance(value, list):
            return YamlTree(value=[YamlTree.wrap(x, span) for x in value], span=span)
        elif isinstance(value, dict):
            return YamlTree(
                value=YamlMap(
                    {
                        YamlTree.wrap(k, span): YamlTree.wrap(v, span)
                        for k, v in value.items()
                    }
                ),
                span=span,
            )
        elif isinstance(value, YamlTree):
            return value
        else:
            return YamlTree(value, span)


class YamlMap:
    """
    To preserve span information for keys, which we commonly use in error messages,
    make a custom map type that is indexable by str, but provides views into all
    necessary spans
    """

    def __init__(self, internal: Dict[YamlTree[str], YamlTree]):
        self._internal = internal

    def __getitem__(self, key: str) -> YamlTree:
        try:
            return next(v for k, v in self._internal.items() if k.value == key)
        except StopIteration:
            raise KeyError(key)

    def __setitem__(self, key: YamlTree[str], value: YamlTree) -> None:
        self._internal[key] = value

    def items(self) -> ItemsView[YamlTree[str], YamlTree]:
        return self._internal.items()

    def key_tree(self, key: str) -> YamlTree[str]:
        return next(k for k, v in self._internal.items() if k.value == key)

    def __contains__(self, item: str) -> bool:
        try:
            _ = self[item]
            return True
        except KeyError:
            return False

    def get(self, key: str) -> Optional[YamlTree]:
        match = [v for k, v in self._internal.items() if k.value == key]
        if match:
            return match[0]
        return None

    def keys(self) -> KeysView[YamlTree[str]]:
        return self._internal.keys()


def parse_yaml_preserve_spans(contents: str, filename: Optional[str]) -> YamlTree:
    """
    parse yaml into a YamlTree object. The resulting spans are tracked in SourceTracker
    so they can be used later when constructing error messages or displaying context.

    :raise jsonschema.exceptions.SchemaError: if config is invalid
    """

    source_hash = SourceTracker.add_source(contents)

    # this uses the `RoundTripConstructor` which inherits from `SafeConstructor`
    class SpanPreservingRuamelConstructor(RoundTripConstructor):
        def construct_object(self, node: Node, deep: bool = False) -> YamlTree:
            r = super().construct_object(node, deep)
            if r is None:
                from semgrep.error import InvalidRuleSchemaError

                Span.from_node(node, source_hash=source_hash, filename=filename)
                raise InvalidRuleSchemaError(
                    short_msg="null values prohibited",
                    long_msg="In semgrep YAML configuration, null values are prohibited",
                    spans=[
                        Span.from_node(
                            node, source_hash=source_hash, filename=filename
                        ).with_context(before=1, after=1)
                    ],
                )

            if isinstance(r, dict):
                r = YamlMap(r)
            return YamlTree(
                r, Span.from_node(node, source_hash=source_hash, filename=filename)
            )

    yaml = YAML()
    yaml.Constructor = SpanPreservingRuamelConstructor
    data = yaml.load(StringIO(contents))

    validate_yaml(data)

    if not isinstance(data, YamlTree):
        raise Exception(
            f"Something went wrong parsing Yaml (expected a YamlTree as output, but got {type(data).__name__}): {PLEASE_FILE_ISSUE_TEXT}"
        )
    return data


class RuleValidation:
    REQUIRE_REGEX = re.compile(r"'(.*)' is a required property")
    PATTERN_KEYS = {
        "pattern",
        "pattern-either",
        "pattern-regex",
        "patterns",
        "pattern-sinks",
        "pattern-sources",
    }
    INVALID_SENTINEL = " is not allowed for "
    BAD_TYPE_SENTINEL = "is not of type"
    BANNED_SENTINEL = "Additional properties are not allowed"


def _validation_error_message(error: jsonschema.exceptions.ValidationError) -> str:  # type: ignore
    """
    Heuristic that returns meaningful error messages in all examples from
    tests/e2e/rules/syntax/badXXX.yaml
    """

    contexts = list(error.parent.context) if error.parent else [error]
    bad_type = set()
    invalid_keys = set()
    any_of_invalid_keys = set()
    required = set()
    banned = set()
    for context in contexts:
        if RuleValidation.BAD_TYPE_SENTINEL in context.message:
            bad_type.add(context.message)
        if RuleValidation.INVALID_SENTINEL in context.message:
            try:
                required_keys = [
                    k["required"][0]
                    for k in context.validator_value.get("anyOf", [])
                    if "required" in k and k["required"]
                ]
                for r in required_keys:
                    if r and r in context.instance.keys():
                        any_of_invalid_keys.add(r)
            except (json.JSONDecodeError, AttributeError):
                invalid_keys.add(context.message)
        if context.message.startswith(RuleValidation.BANNED_SENTINEL):
            banned.add(context.message)
        require_matches = RuleValidation.REQUIRE_REGEX.match(context.message)
        if require_matches:
            required.add(require_matches[1])

    if invalid_keys:
        return "\n".join(sorted(invalid_keys))
    if bad_type:
        return "\n".join(sorted(bad_type))
    if banned:
        return "\n".join(sorted(banned))

    outs = []
    if any_of_invalid_keys:
        keys = ", ".join(f"'{k}'" for k in sorted(any_of_invalid_keys))
        outs.append(f"One of these properties may be invalid: {keys}")
        required = required - RuleValidation.PATTERN_KEYS
    if required:
        keys = ", ".join(f"'{k}'" for k in sorted(required))
        outs.append(f"One of these properties is missing: {keys}")
    if outs:
        return "\n".join(outs)

    return cast(str, contexts[0].message)


def validate_yaml(data: YamlTree) -> None:
    from semgrep.error import InvalidRuleSchemaError

    try:
        jsonschema.validate(data.unroll(), RuleSchema.get(), cls=Draft7Validator)
    except jsonschema.ValidationError as ve:
        message = _validation_error_message(ve)
        item = data

        root_error = ve
        while root_error.parent is not None:
            root_error = root_error.parent

        for el in root_error.absolute_path:
            item = item.value[el]

        raise InvalidRuleSchemaError(
            short_msg="Invalid rule schema",
            long_msg=message,
            spans=[item.span],
        )



from typing import Any
from typing import Dict
from typing import Optional

from semgrep.rule_lang import Span
from semgrep.semgrep_types import BooleanRuleExpression
from semgrep.semgrep_types import Language


class Pattern:
    """
    Subunit of Rule that semgrep-core runs on
    """

    def __init__(
        self,
        rule_index: int,
        expression: BooleanRuleExpression,
        severity: str,
        language: Language,
        span: Optional[Span],
    ) -> None:
        self._id = f"{rule_index}.{expression.pattern_id}"
        self._language = language
        self._severity = severity
        self._expression = expression
        self._pattern = expression.operand
        self._span = span

    @property
    def span(self) -> Optional[Span]:
        return self._span

    @property
    def language(self) -> Language:
        return self._language

    @property
    def expression(self) -> BooleanRuleExpression:
        return self._expression

    def to_json(self) -> Dict[str, Any]:
        # Note languages is a list as this is what semgrep-core expects
        return {
            "id": self._id,
            "pattern": self._pattern,
            "severity": self._severity,
            "languages": [self._language],
            "message": "<internalonly>",
        }

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self._id}>"



import subprocess
import sys
from pathlib import Path

from semgrep.semgrep_main import invoke_semgrep


def test_api(capsys, run_semgrep_in_tmp):
    # Test that exposed python API works and prints out nothing to stderr or stdout
    output = invoke_semgrep(
        Path("rules/eqeq.yaml"),
        [Path("targets/bad/invalid_python.py"), Path("targets/basic/stupid.py")],
    )

    captured = capsys.readouterr()
    assert len(output["errors"]) == 1
    assert len(output["results"]) == 1
    assert captured.out == ""
    assert captured.err == ""

    # Check that logging code isnt handled by default root handler and printed to stderr
    x = subprocess.run(
        [
            sys.executable,
            "-c",
            "from semgrep.semgrep_main import invoke_semgrep; from pathlib import Path; invoke_semgrep(Path('rules/eqeq.yaml'),[Path('targets/bad/invalid_python.py'), Path('targets/basic/stupid.py')],)",
        ],
        encoding="utf-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    assert x.stdout == ""
    assert x.stderr == ""



import functools
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import NamedTuple
from typing import NewType
from typing import Optional
from typing import Union

import attr

from semgrep.rule_lang import YamlMap


Language = NewType("Language", str)
Mode = NewType("Mode", str)
PatternId = NewType("PatternId", str)
Operator = NewType("Operator", str)
FileExtension = NewType("FileExtension", str)

TAINT_MODE = Mode("taint")
SEARCH_MODE = DEFAULT_MODE = Mode("search")
SUPPORTED_MODES = {TAINT_MODE, SEARCH_MODE}

YAML_TAINT_MUST_HAVE_KEYS = {"pattern-sinks", "pattern-sources"}


class OPERATORS:
    AND_ALL: Operator = Operator("and_all")
    AND_NOT: Operator = Operator("and_not")
    AND: Operator = Operator("and")
    AND_EITHER: Operator = Operator("and_either")
    AND_INSIDE: Operator = Operator("and_inside")
    AND_NOT_INSIDE: Operator = Operator("and_not_inside")
    WHERE_PYTHON: Operator = Operator("where_python")
    FIX: Operator = Operator("fix")
    FIX_REGEX: Operator = Operator("fix_regex")
    EQUIVALENCES: Operator = Operator("equivalences")
    REGEX: Operator = Operator("regex")
    NOT_REGEX: Operator = Operator("not_regex")
    METAVARIABLE_REGEX: Operator = Operator("metavariable_regex")
    METAVARIABLE_COMPARISON: Operator = Operator("metavariable_comparison")


OPERATORS_WITH_CHILDREN = [OPERATORS.AND_ALL, OPERATORS.AND_EITHER]

OPERATOR_PATTERN_NAMES_MAP = {
    OPERATORS.AND_INSIDE: ["pattern-inside"],
    OPERATORS.AND_NOT_INSIDE: ["pattern-not-inside"],
    OPERATORS.AND_EITHER: ["pattern-either"],
    OPERATORS.AND_NOT: ["pattern-not"],
    OPERATORS.AND: ["pattern"],
    OPERATORS.AND_ALL: ["patterns"],
    OPERATORS.WHERE_PYTHON: ["pattern-where-python"],
    OPERATORS.FIX: ["fix"],
    OPERATORS.FIX_REGEX: ["fix-regex"],
    OPERATORS.EQUIVALENCES: ["equivalences"],
    OPERATORS.REGEX: ["pattern-regex"],
    OPERATORS.NOT_REGEX: ["pattern-not-regex"],
    OPERATORS.METAVARIABLE_REGEX: ["metavariable-regex"],
    OPERATORS.METAVARIABLE_COMPARISON: ["metavariable-comparison"],
}
PATTERN_NAMES_OPERATOR_MAP = {
    v: k for k, vv in OPERATOR_PATTERN_NAMES_MAP.items() for v in vv
}


@attr.s(auto_attribs=True, frozen=True)
class BooleanRuleExpression:
    operator: Operator
    pattern_id: Optional[PatternId] = None
    children: Optional[List["BooleanRuleExpression"]] = None
    operand: Optional[Union[str, YamlMap]] = None


def pattern_name_for_operator(operator: Operator) -> str:
    # TODO
    return pattern_names_for_operator(operator)[0]


def pattern_names_for_operator(operator: Operator) -> List[str]:
    return OPERATOR_PATTERN_NAMES_MAP[operator]


def pattern_names_for_operators(operators: List[Operator]) -> List[str]:
    return sum((pattern_names_for_operator(op) for op in operators), [])


ALLOWED_GLOB_TYPES = ("include", "exclude")


@functools.total_ordering
class Range(NamedTuple):
    start: int
    end: int
    metavariables: Mapping[str, Any]
    propagated_metavariables: Dict[str, Any] = {}

    def is_enclosing_or_eq(self, rhs: "Range") -> bool:
        return self.is_range_enclosing_or_eq(rhs) and self.metavariables_match(rhs)

    def is_range_enclosing_or_eq(self, rhs: "Range") -> bool:
        return self.start <= rhs.start and rhs.end <= self.end

    def metavariables_match(self, rhs: "Range") -> bool:
        """
        Returns true if and only if all metavariables in both this and the other Range
        refer to the same variables (if variable nodes), in the same scope, or the same
        expressions (if expression nodes).

        That is, if two patterns define a "$X", and $X refers to a variable in one
        pattern, then $X must refer to the same variable in both patterns.
        """
        to_match = set(self.metavariables.keys()) & set(rhs.metavariables.keys())
        return all(
            self.metavariables[v]
            and rhs.metavariables[v]
            and self.metavariables[v] == rhs.metavariables[v]
            for v in to_match
        )

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f"start={self.start} "
            f"end={self.end} "
            f"metavariables={self.metavariables} "
            f"propagated={self.propagated_metavariables}>"
        )

    def __hash__(self) -> int:
        return hash((self.start, self.end))

    def __eq__(self, rhs: Any) -> bool:
        if not isinstance(rhs, type(self)):
            return False

        return (
            self.start == rhs.start
            and self.end == rhs.end
            and self.metavariables_match(rhs)
        )

    def __lt__(self, rhs: Any) -> bool:
        if not isinstance(rhs, type(self)):
            return False

        diff_self = self.end - self.start
        diff_rhs = rhs.end - rhs.start

        if diff_self == diff_rhs:
            return self.start < rhs.start

        return diff_self < diff_rhs



import json
import subprocess


def test_synthesize_patterns():
    output = subprocess.check_output(
        [
            "python",
            "-m",
            "semgrep",
            "--synthesize-patterns",
            "6:10-6:30",
            "tests/e2e/rules/synthesizing/ex1.py",
        ],
        encoding="utf-8",
    )
    print("output = ", output)
    output_json = json.loads(output)
    assert output_json["exact match"] == "metrics.send('my-report-id')"
    assert output_json["dots"] == "metrics.send(...)"
